# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import onnx
import onnx.numpy_helper as np_helper
import os
import urllib.request as ureq
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.general import GiveReadableTensorNames, GiveUniqueNodeNames
from qonnx.transformation.infer_shapes import InferShapes


def test_renaming():
    # load the onnx model
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    # do some basic checks
    assert model.graph.input[0].name == "global_in"
    assert model.graph.output[0].name == "global_out"
    assert model.graph.node[1].op_type == "Conv"
    assert model.graph.node[1].name == "Conv_0"
    assert model.graph.node[1].input[1] == "Conv_0_param0"
    assert model.graph.node[6].op_type == "Add"
    assert model.graph.node[6].name == "Add_1"
    assert model.graph.node[6].input[1] == "Add_1_param0"
    # ensure running renaming twice still yields the same names
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    assert model.graph.node[1].op_type == "Conv"
    assert model.graph.node[1].name == "Conv_0"
    assert model.graph.node[1].input[1] == "Conv_0_param0"
    assert model.graph.node[6].op_type == "Add"
    assert model.graph.node[6].name == "Add_1"
    assert model.graph.node[6].input[1] == "Add_1_param0"
    # run renamed model to make sure we did not mess up the topology
    raw_i = get_data("qonnx.data", "onnx/mnist-conv/test_data_set_0/input_0.pb")
    raw_o = get_data("qonnx.data", "onnx/mnist-conv/test_data_set_0/output_0.pb")
    input_tensor = onnx.load_tensor_from_string(raw_i)
    output_tensor = onnx.load_tensor_from_string(raw_o)
    input_dict = {"global_in": np_helper.to_array(input_tensor)}
    output_dict = oxe.execute_onnx(model, input_dict)
    assert np.isclose(np_helper.to_array(output_tensor), output_dict["global_out"], atol=1e-3).all()


def test_rename_multi_io_tinyyolov3():
    download_url = "https://github.com/onnx/models/raw/main/validated/vision/object_detection_segmentation"
    download_url += "/tiny-yolov3/model/tiny-yolov3-11.onnx"
    export_onnx_path = download_url.split("/")[-1]
    ureq.urlretrieve(download_url, export_onnx_path)
    if not os.path.isfile(export_onnx_path):
        pytest.skip("Couldn't download ONNX model, skipping")
    model = ModelWrapper(export_onnx_path)
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    assert len(model.graph.input) == 2
    assert model.graph.input[0].name == "global_in"
    assert model.graph.input[1].name == "global_in_1"
    assert len(model.graph.output) == 3
    assert model.graph.output[0].name == "global_out"
    assert model.graph.output[1].name == "global_out_1"
    assert model.graph.output[2].name == "global_out_2"
    os.remove(export_onnx_path)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.pruning import PruneChannels


def prune_channels(input_filename, prunespec_filename, *, lossy=True, output_filename=""):
    """
    Prune channels from specified tensors and their dependencies from a model.
    The model must have already been cleaned up by qonnx-cleanup, including the
    --extract-conv-bias=True --preserve-qnt-ops=False options.

    :param input_filename: Filename for the input ONNX model
    :param prunespec_filename: Filename for the pruning specification, formatted as a Python dict
        formatted as {tensor_name : {axis : {channels}}}. See test_pruning.py for examples.
    :param lossy: Whether to perform lossy pruning, see the PruneChannels transformation for description.
    :param output_filename: If specified, write the resulting pruned model to this filename. Otherwise,
        the input_filename will be used with a _pruned suffix.
    """
    model = ModelWrapper(input_filename)
    with open(prunespec_filename) as f:
        prunespec_dict = dict(eval(f.read()))
    pruned_model = model.transform(PruneChannels(prunespec_dict, lossy))
    if output_filename == "":
        output_filename = input_filename.replace(".onnx", "_pruned.onnx")
    pruned_model.save(output_filename)


def main():
    clize.run(prune_channels)


if __name__ == "__main__":
    main()



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto, helper

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation
from qonnx.transformation.general import GiveUniqueParameterTensors, SortGraph
from qonnx.transformation.remove import RemoveIdentityOps


class ExtractQuantScaleZeroPt(Transformation):
    """Extract any non-identity scale and zero-point Quant inputs as
    separate Div/Mul (for scale) and Add/Sub (for zeropoint" nodes,
    preceding and following the Quant node."""

    def apply(self, model: ModelWrapper):
        graph = model.graph
        for node in graph.node:
            if node.op_type == "Quant":
                quant_node = node
                input_nm, scale_nm, zeropt_nm, _ = node.input
                scale_t = model.get_initializer(scale_nm)
                zeropt_t = model.get_initializer(zeropt_nm)
                ishp = model.get_tensor_shape(input_nm)
                extract_scale = False
                extract_zeropt = False
                if scale_t is not None and (scale_t != 1).any():
                    extract_scale = True
                if zeropt_t is not None and (zeropt_t != 0).any():
                    extract_zeropt = True
                if (not extract_scale) and (not extract_zeropt):
                    continue
                running_input = input_nm
                if extract_scale:
                    # create new Div node that divides the input
                    # by the scale
                    inp_scaled_nm = model.make_new_valueinfo_name()
                    inp_scaled = helper.make_tensor_value_info(
                        inp_scaled_nm,
                        TensorProto.FLOAT,
                        ishp,
                    )
                    graph.value_info.append(inp_scaled)
                    inp_scale_node = helper.make_node("Div", [running_input, scale_nm], [inp_scaled_nm])
                    graph.node.append(inp_scale_node)
                    # create new Mul node
                    # remove scale from Quant node
                    new_scale_nm = model.make_new_valueinfo_name()
                    model.set_initializer(new_scale_nm, np.asarray(1.0, dtype=np.float32))
                    quant_node.input[1] = new_scale_nm
                    running_input = inp_scaled_nm
                if extract_zeropt:
                    # create new Add node that adds the zeropoint to
                    # the scaled input
                    inp_zeropt_nm = model.make_new_valueinfo_name()
                    inp_zeropt = helper.make_tensor_value_info(
                        inp_zeropt_nm,
                        TensorProto.FLOAT,
                        ishp,
                    )
                    graph.value_info.append(inp_zeropt)
                    inp_zeropt_node = helper.make_node("Add", [running_input, zeropt_nm], [inp_zeropt_nm])
                    graph.node.append(inp_zeropt_node)
                    # remove zeropt from Quant node
                    new_zeropt_nm = model.make_new_valueinfo_name()
                    model.set_initializer(new_zeropt_nm, np.asarray(0.0, dtype=np.float32))
                    quant_node.input[2] = new_zeropt_nm
                    running_input = inp_zeropt_nm
                # rewire node input to any newly created Div/Add nodes
                quant_node.input[0] = running_input
                last_node = quant_node
                final_output = quant_node.output[0]
                if extract_zeropt:
                    # create new Sub node that subtracts the zeropoint from
                    # the output
                    out_zeropt_nm = model.make_new_valueinfo_name()
                    out_zeropt = helper.make_tensor_value_info(
                        out_zeropt_nm,
                        TensorProto.FLOAT,
                        ishp,
                    )
                    graph.value_info.append(out_zeropt)
                    out_zeropt_node = helper.make_node("Sub", [out_zeropt_nm, zeropt_nm], [final_output])
                    last_node.output[0] = out_zeropt_nm
                    graph.node.append(out_zeropt_node)
                    # important: when tracking a pointer to newly added nodes,
                    # ensure the item from the container is used, and not the
                    # make_node result -- those are different objects
                    # e.g. if we use last_node = out_zeropt_node below,
                    # this will point to the wrong object and cause bugs later
                    last_node = graph.node[-1]
                if extract_scale:
                    # create new Mul node that applies the output scale
                    out_scale_nm = model.make_new_valueinfo_name()
                    out_scale = helper.make_tensor_value_info(
                        out_scale_nm,
                        TensorProto.FLOAT,
                        ishp,
                    )
                    last_node.output[0] = out_scale_nm
                    graph.value_info.append(out_scale)
                    out_scale_node = helper.make_node("Mul", [out_scale_nm, scale_nm], [final_output])
                    graph.node.append(out_scale_node)

                if extract_scale or extract_zeropt:
                    # since we used append() for new nodes, need to call
                    # SortGraph to ensure correct (topological) order
                    model = model.transform(SortGraph())
                    # Remove potential unity multiplications from alpha and beta attributes
                    model = model.transform(RemoveIdentityOps())
                    # Ensure unique parameter tensors
                    model = model.transform(GiveUniqueParameterTensors())
                    return model, True

        return model, False



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
from onnx import TensorProto
from onnx import helper as oh

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.create_generic_partitions import PartitionFromDict
from qonnx.transformation.extend_partition import ExtendPartition
from qonnx.util.basic import gen_finn_dt_tensor, qonnx_make_model


def create_model():
    MultiThreshold0_node = oh.make_node(
        "MultiThreshold",
        inputs=["in1_multithreshold0", "in2_multithreshold0"],
        outputs=["out_multithreshold0"],
        name="MultiThreshold0",
        domain="qonnx.custom_op.general",
        out_dtype="UINT4",
    )

    Conv0_node = oh.make_node(
        "Conv",
        inputs=["out_multithreshold0", "in2_conv0"],
        outputs=["out_conv0"],
        name="Conv0",
        dilations=[1, 1],
        group=1,
        kernel_shape=[1, 1],
        pads=[0, 0, 0, 0],
        strides=[1, 1],
    )

    Conv1_node = oh.make_node(
        "Conv",
        inputs=["out_multithreshold0", "in2_conv1"],
        outputs=["out_conv1"],
        name="Conv1",
        dilations=[1, 1],
        group=1,
        kernel_shape=[1, 1],
        pads=[0, 0, 0, 0],
        strides=[1, 1],
    )

    MultiThreshold1_node = oh.make_node(
        "MultiThreshold",
        inputs=["out_conv0", "in2_multithreshold1"],
        outputs=["out_multithreshold1"],
        name="MultiThreshold1",
        domain="qonnx.custom_op.general",
        out_dtype="UINT4",
    )

    MultiThreshold2_node = oh.make_node(
        "MultiThreshold",
        inputs=["out_conv1", "in2_multithreshold2"],
        outputs=["out_multithreshold2"],
        name="MultiThreshold2",
        domain="qonnx.custom_op.general",
        out_dtype="UINT4",
    )

    Add0_node = oh.make_node(
        "Add",
        inputs=["out_multithreshold1", "out_multithreshold2"],
        outputs=["out_add0"],
        name="Add0",
    )

    MultiThreshold3_node = oh.make_node(
        "MultiThreshold",
        inputs=["out_add0", "in2_multithreshold3"],
        outputs=["out_multithreshold3"],
        name="MultiThreshold3",
        domain="qonnx.custom_op.general",
        out_dtype="UINT4",
    )

    Conv2_node = oh.make_node(
        "Conv",
        inputs=["out_multithreshold3", "in2_conv2"],
        outputs=["out_conv2"],
        name="Conv2",
        dilations=[1, 1],
        group=1,
        kernel_shape=[1, 1],
        pads=[0, 0, 0, 0],
        strides=[1, 1],
    )

    Conv3_node = oh.make_node(
        "Conv",
        inputs=["out_multithreshold3", "in2_conv3"],
        outputs=["out_conv3"],
        name="Conv3",
        dilations=[1, 1],
        group=1,
        kernel_shape=[1, 1],
        pads=[0, 0, 0, 0],
        strides=[1, 1],
    )

    MultiThreshold4_node = oh.make_node(
        "MultiThreshold",
        inputs=["out_conv2", "in2_multithreshold4"],
        outputs=["out_multithreshold4"],
        name="MultiThreshold4",
        domain="qonnx.custom_op.general",
        out_dtype="UINT4",
    )

    MultiThreshold5_node = oh.make_node(
        "MultiThreshold",
        inputs=["out_conv3", "in2_multithreshold5"],
        outputs=["out_multithreshold5"],
        name="MultiThreshold5",
        domain="qonnx.custom_op.general",
        out_dtype="UINT4",
    )

    Add1_node = oh.make_node(
        "Add",
        inputs=["out_multithreshold4", "out_multithreshold5"],
        outputs=["out_add1"],
        name="Add1",
    )

    # Inputs/outputs (global)
    t_type = TensorProto.FLOAT
    t_shape = [1, 256, 128, 1]
    in1_multithreshold0 = oh.make_tensor_value_info("in1_multithreshold0", t_type, t_shape)
    out_add1 = oh.make_tensor_value_info("out_add1", t_type, t_shape)

    # Initializers
    in2_multithreshold0 = oh.make_tensor_value_info("in2_multithreshold0", t_type, [256, 15])
    in2_conv0 = oh.make_tensor_value_info("in2_conv0", t_type, [256, 256, 1, 1])
    in2_conv1 = oh.make_tensor_value_info("in2_conv1", t_type, [256, 256, 1, 1])
    in2_multithreshold1 = oh.make_tensor_value_info("in2_multithreshold1", t_type, [256, 15])
    in2_multithreshold2 = oh.make_tensor_value_info("in2_multithreshold2", t_type, [256, 15])
    in2_multithreshold3 = oh.make_tensor_value_info("in2_multithreshold3", t_type, [256, 15])
    in2_conv2 = oh.make_tensor_value_info("in2_conv2", t_type, [256, 256, 1, 1])
    in2_conv3 = oh.make_tensor_value_info("in2_conv3", t_type, [256, 256, 1, 1])
    in2_multithreshold4 = oh.make_tensor_value_info("in2_multithreshold4", t_type, [256, 15])
    in2_multithreshold5 = oh.make_tensor_value_info("in2_multithreshold5", t_type, [256, 15])

    # Value_infos
    out_multithreshold0 = oh.make_tensor_value_info("out_multithreshold0", t_type, t_shape)
    out_conv0 = oh.make_tensor_value_info("out_conv0", t_type, t_shape)
    out_conv1 = oh.make_tensor_value_info("out_conv1", t_type, t_shape)
    out_multithreshold1 = oh.make_tensor_value_info("out_multithreshold1", t_type, t_shape)
    out_multithreshold2 = oh.make_tensor_value_info("out_multithreshold2", t_type, t_shape)
    out_add0 = oh.make_tensor_value_info("out_add0", t_type, t_shape)
    out_multithreshold3 = oh.make_tensor_value_info("out_multithreshold3", t_type, t_shape)
    out_conv2 = oh.make_tensor_value_info("out_conv2", t_type, t_shape)
    out_conv3 = oh.make_tensor_value_info("out_conv3", t_type, t_shape)
    out_multithreshold4 = oh.make_tensor_value_info("out_multithreshold4", t_type, t_shape)
    out_multithreshold5 = oh.make_tensor_value_info("out_multithreshold5", t_type, t_shape)

    graph = oh.make_graph(
        nodes=[
            MultiThreshold0_node,
            Conv0_node,
            Conv1_node,
            MultiThreshold1_node,
            MultiThreshold2_node,
            Add0_node,
            MultiThreshold3_node,
            Conv2_node,
            Conv3_node,
            MultiThreshold4_node,
            MultiThreshold5_node,
            Add1_node,
        ],
        name="test_graph",
        inputs=[in1_multithreshold0],
        outputs=[out_add1],
        value_info=[
            in2_multithreshold0,
            in2_conv0,
            in2_conv1,
            in2_multithreshold1,
            in2_multithreshold2,
            in2_multithreshold3,
            in2_conv2,
            in2_conv3,
            in2_multithreshold4,
            in2_multithreshold5,
            out_multithreshold0,
            out_conv0,
            out_conv1,
            out_multithreshold1,
            out_multithreshold2,
            out_add0,
            out_multithreshold3,
            out_conv2,
            out_conv3,
            out_multithreshold4,
            out_multithreshold5,
        ],
    )

    onnx_model = qonnx_make_model(graph, producer_name="test_model")
    model = ModelWrapper(onnx_model)

    mt_weights = np.random.randint(low=-1000, high=1000, size=[6, 256, 15])
    mt_weights = np.sort(mt_weights, 2)
    for i in range(0, 6):
        model.set_initializer("in2_multithreshold" + str(i), mt_weights[i])

    conv_weights = np.random.randint(low=-8, high=7, size=[4, 256, 256, 1, 1]).astype(np.float32)
    for i in range(0, 4):
        model.set_initializer("in2_conv" + str(i), conv_weights[i])
        model.set_tensor_datatype("in2_conv" + str(i), DataType["INT4"])

    return model


# Partitioning
@pytest.mark.parametrize("p", [0, 1, 2])
# Extending
@pytest.mark.parametrize("extend_id", [[0], [1], [0, 1]])
def test_extend_partition(p, extend_id):
    if p == 0:
        if extend_id != [0]:
            pytest.skip("Only the first partition node can be extended")
    if p == 1:
        if extend_id != [1]:
            pytest.skip("Only the second partition node can be extended")
        else:
            extend_id = [6]  # The 6th node is the index of the GenericPartition
            # node, so we set the index to the right value

    model = create_model()

    # Partition the model first
    partitionings = [
        {0: range(0, 6)},
        {0: range(6, 12)},
        {0: range(0, 6), 1: range(6, 12)},
    ]
    partitioning = partitionings[p]

    model = model.transform(PartitionFromDict(partitioning))

    # Create input data
    input0_tensor_name = model.graph.input[0].name

    input_shape = model.get_tensor_shape(input0_tensor_name)
    input_dtype = model.get_tensor_datatype(input0_tensor_name)
    input_val = gen_finn_dt_tensor(input_dtype, input_shape)
    input_dict = {}
    input_dict[input0_tensor_name] = input_val

    # Extend the model
    model_extended = model.transform(ExtendPartition(extend_id))

    assert oxe.compare_execution(model, model_extended, input_dict)

    # Check if FINN data_types are retained
    for n in model_extended.graph.node:
        if n.op_type == "Conv":
            assert model_extended.get_tensor_datatype(n.input[1]) == DataType["INT4"]



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.qcdq_to_qonnx import QCDQToQuant
from qonnx.transformation.qonnx_to_qcdq import QuantToQCDQ

CONVERT_MODE_QCDQ = "qcdq"
CONVERT_MODE_QUANT = "quant"

convert_modes = {CONVERT_MODE_QCDQ, CONVERT_MODE_QUANT}

convert_mode_options = clize.parameters.mapped(
    [
        (CONVERT_MODE_QCDQ, [CONVERT_MODE_QCDQ], "Convert from Quant to QCDQ"),
        (CONVERT_MODE_QUANT, [CONVERT_MODE_QUANT], "Convert from QCDQ to Quant"),
    ]
)


def convert(input_model_file, *, output_style: convert_mode_options, output_file: str = None):
    """Convert an ONNX file from one style of quantization to another, where possible.
    Please see the documentation on the QuantToQCDQ and QCDQToQuant
    transformations to learn more about the particular limitations.

    :param input_model_file: Filename for the input ONNX model.
    :param output_style: Quantization style for the output.
    :param output_file: If specified, write the output ONNX model to this filename.
        Otherwise, will default to the input file with an _output_style suffix.
    """
    model = ModelWrapper(input_model_file)
    if output_style == CONVERT_MODE_QCDQ:
        model = model.transform(QuantToQCDQ())
    elif output_style == CONVERT_MODE_QUANT:
        model = model.transform(QCDQToQuant())
    else:
        print("Unknown output_style for conversion: %s" % output_style)
        exit(-1)
    if output_file is None:
        output_file = input_model_file.replace(".onnx", "_%s.onnx" % output_style)
    model.save(output_file)


def main():
    clize.run(convert)


if __name__ == "__main__":
    main()



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np

import qonnx.core.onnx_exec as oxe
from qonnx.transformation.change_batchsize import ChangeBatchSize
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.onnx import valueinfo_to_tensor
from qonnx.util.test import download_model, test_model_details

model_details = test_model_details


@pytest.mark.parametrize("test_model", model_details.keys())
def test_change_batchsize(test_model):
    test_details = model_details[test_model]
    batch_size = 10
    old_ishape = test_details["input_shape"]
    imin, imax = test_details["input_range"]
    # some models spec per-channel ranges, be conservative for those
    if isinstance(imin, np.ndarray):
        imin = imin.max()
    if isinstance(imax, np.ndarray):
        imax = imax.min()
    model = download_model(test_model=test_model, do_cleanup=True, return_modelwrapper=True)
    iname = model.graph.input[0].name
    oname = model.graph.output[0].name
    example_inp = valueinfo_to_tensor(model.get_tensor_valueinfo(iname))
    assert tuple(model.get_tensor_shape(iname)) == old_ishape
    model = model.transform(ChangeBatchSize(batch_size))
    model = model.transform(InferShapes())
    exp_ishape = (batch_size, *old_ishape[1:])
    assert tuple(model.get_tensor_shape(iname)) == exp_ishape
    new_inp = np.random.uniform(imin, imax, exp_ishape).astype(example_inp.dtype)
    ret = oxe.execute_onnx(model, {iname: new_inp})
    assert ret[oname].shape[0] == batch_size



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import numpy as np
from abc import ABC, abstractmethod
from enum import Enum, EnumMeta


class BaseDataType(ABC):
    "Base class for QONNX data types."

    def signed(self):
        "Returns whether this DataType can represent negative numbers."
        return self.min() < 0

    def __eq__(self, other):
        if isinstance(other, BaseDataType):
            return self.get_canonical_name() == other.get_canonical_name()
        elif isinstance(other, str):
            return self.get_canonical_name() == other
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.get_canonical_name())

    @property
    def name(self):
        return self.get_canonical_name()

    def __repr__(self):
        return self.get_canonical_name()

    def __str__(self):
        return self.get_canonical_name()

    @abstractmethod
    def bitwidth(self):
        "Returns the number of bits required for this DataType."
        pass

    @abstractmethod
    def min(self):
        "Returns the smallest possible value allowed by this DataType."
        pass

    @abstractmethod
    def max(self):
        "Returns the largest possible value allowed by this DataType."
        pass

    @abstractmethod
    def allowed(self, value):
        """Check whether given value is allowed for this DataType.

        * value (float32): value to be checked"""
        pass

    @abstractmethod
    def get_num_possible_values(self):
        """Returns the number of possible values this DataType can take. Only
        implemented for integer types for now."""
        pass

    @abstractmethod
    def is_integer(self):
        "Returns whether this DataType represents integer values only."
        pass

    @abstractmethod
    def is_fixed_point(self):
        "Returns whether this DataType represent fixed-point values only."
        pass

    @abstractmethod
    def get_hls_datatype_str(self):
        "Returns the corresponding Vivado HLS datatype name."
        pass

    @abstractmethod
    def to_numpy_dt(self):
        "Return an appropriate numpy datatype that can represent this QONNX DataType."
        pass

    @abstractmethod
    def get_canonical_name(self):
        "Return a canonical string representation of this QONNX DataType."


class FloatType(BaseDataType):
    def bitwidth(self):
        return 32

    def min(self):
        return np.finfo(np.float32).min

    def max(self):
        return np.finfo(np.float32).max

    def allowed(self, value):
        return True

    def get_num_possible_values(self):
        raise Exception("Undefined for FloatType")

    def is_integer(self):
        return False

    def is_fixed_point(self):
        return False

    def get_hls_datatype_str(self):
        return "float"

    def to_numpy_dt(self):
        return np.float32

    def get_canonical_name(self):
        return "FLOAT32"


class Float16Type(BaseDataType):
    def bitwidth(self):
        return 16

    def min(self):
        return np.finfo(np.float16).min

    def max(self):
        return np.finfo(np.float16).max

    def allowed(self, value):
        return True

    def get_num_possible_values(self):
        raise Exception("Undefined for Float16Type")

    def is_integer(self):
        return False

    def is_fixed_point(self):
        return False

    def get_hls_datatype_str(self):
        return "half"

    def to_numpy_dt(self):
        return np.float16

    def get_canonical_name(self):
        return "FLOAT16"


class IntType(BaseDataType):
    def __init__(self, bitwidth, signed):
        super().__init__()
        self._bitwidth = bitwidth
        self._signed = signed

    def bitwidth(self):
        return self._bitwidth

    def min(self):
        unsigned_min = 0
        signed_min = -(2 ** (self.bitwidth() - 1))
        return signed_min if self._signed else unsigned_min

    def max(self):
        unsigned_max = (2 ** (self.bitwidth())) - 1
        signed_max = (2 ** (self.bitwidth() - 1)) - 1
        return signed_max if self._signed else unsigned_max

    def allowed(self, value):
        return (self.min() <= value) and (value <= self.max()) and float(value).is_integer()

    def get_num_possible_values(self):
        return abs(self.min()) + abs(self.max()) + 1

    def is_integer(self):
        return True

    def is_fixed_point(self):
        return False

    def get_hls_datatype_str(self):
        if self.signed():
            return "ap_int<%d>" % self.bitwidth()
        else:
            return "ap_uint<%d>" % self.bitwidth()

    def to_numpy_dt(self):
        if self.bitwidth() <= 8:
            return np.int8 if self.signed() else np.uint8
        elif self.bitwidth() <= 16:
            return np.int16 if self.signed() else np.uint16
        elif self.bitwidth() <= 32:
            return np.int32 if self.signed() else np.uint32
        elif self.bitwidth() <= 64:
            return np.int64 if self.signed() else np.uint64
        else:
            raise Exception("Unknown numpy dtype for " + str(self))

    def get_canonical_name(self):
        if self.bitwidth() == 1 and (not self.signed()):
            return "BINARY"
        else:
            prefix = "INT" if self.signed() else "UINT"
            return prefix + str(self.bitwidth())


class BipolarType(BaseDataType):
    def bitwidth(self):
        return 1

    def min(self):
        return -1

    def max(self):
        return +1

    def allowed(self, value):
        return value in [-1, +1]

    def get_num_possible_values(self):
        return 2

    def is_integer(self):
        return True

    def is_fixed_point(self):
        return False

    def get_hls_datatype_str(self):
        return "ap_int<1>"

    def to_numpy_dt(self):
        return np.int8

    def get_canonical_name(self):
        return "BIPOLAR"


class TernaryType(BaseDataType):
    def bitwidth(self):
        return 2

    def min(self):
        return -1

    def max(self):
        return +1

    def allowed(self, value):
        return value in [-1, 0, +1]

    def get_num_possible_values(self):
        return 3

    def is_integer(self):
        return True

    def is_fixed_point(self):
        return False

    def get_hls_datatype_str(self):
        return "ap_int<2>"

    def to_numpy_dt(self):
        return np.int8

    def get_canonical_name(self):
        return "TERNARY"


class FixedPointType(IntType):
    def __init__(self, bitwidth, intwidth):
        super().__init__(bitwidth=bitwidth, signed=True)
        assert intwidth < bitwidth, "FixedPointType violates intwidth < bitwidth"
        self._intwidth = intwidth

    def int_bits(self):
        return self._intwidth

    def frac_bits(self):
        return self.bitwidth() - self.int_bits()

    def scale_factor(self):
        return 2 ** -(self.frac_bits())

    def min(self):
        return super().min() * self.scale_factor()

    def max(self):
        return super().max() * self.scale_factor()

    def allowed(self, value):
        int_value = value / self.scale_factor()
        return IntType(self._bitwidth, True).allowed(int_value)

    def is_integer(self):
        return False

    def is_fixed_point(self):
        return True

    def get_hls_datatype_str(self):
        return "ap_fixed<%d, %d>" % (self.bitwidth(), self.int_bits())

    def to_numpy_dt(self):
        return np.float32

    def get_canonical_name(self):
        return "FIXED<%d,%d>" % (self.bitwidth(), self.int_bits())


class ScaledIntType(IntType):
    # scaled integer datatype, only intended for
    # inference cost calculations, many of the
    # member methods are not implemented
    def __init__(self, bitwidth):
        super().__init__(bitwidth=bitwidth, signed=True)

    def min(self):
        raise Exception("Undefined for ScaledIntType")

    def max(self):
        raise Exception("Undefined for ScaledIntType")

    def allowed(self, value):
        raise Exception("Undefined for ScaledIntType")

    def is_integer(self):
        return False

    def is_fixed_point(self):
        return False

    def get_hls_datatype_str(self):
        raise Exception("Undefined for ScaledIntType")

    def to_numpy_dt(self):
        return np.float32

    def signed(self):
        "Returns whether this DataType can represent negative numbers."
        return True

    def get_canonical_name(self):
        return "SCALEDINT<%d>" % (self.bitwidth())


def resolve_datatype(name):
    if not isinstance(name, str):
        raise TypeError(f"Input 'name' must be of type 'str', but got type '{type(name).__name__}'")

    _special_types = {
        "BINARY": IntType(1, False),
        "BIPOLAR": BipolarType(),
        "TERNARY": TernaryType(),
        "FLOAT32": FloatType(),
        "FLOAT16": Float16Type(),
    }
    if name in _special_types.keys():
        return _special_types[name]
    elif name.startswith("UINT"):
        bitwidth = int(name.replace("UINT", ""))
        return IntType(bitwidth, False)
    elif name.startswith("INT"):
        bitwidth = int(name.replace("INT", ""))
        return IntType(bitwidth, True)
    elif name.startswith("FIXED"):
        name = name.replace("FIXED<", "")
        name = name.replace(">", "")
        nums = name.split(",")
        bitwidth = int(nums[0].strip())
        intwidth = int(nums[1].strip())
        return FixedPointType(bitwidth, intwidth)
    elif name.startswith("SCALEDINT"):
        name = name.replace("SCALEDINT<", "")
        name = name.replace(">", "")
        nums = name.split(",")
        bitwidth = int(nums[0].strip())
        return ScaledIntType(bitwidth)
    else:
        raise KeyError("Could not resolve DataType " + name)


class DataTypeMeta(EnumMeta):
    def __getitem__(self, name):
        return resolve_datatype(name)


class DataType(Enum, metaclass=DataTypeMeta):
    """Enum class that contains QONNX data types to set the quantization annotation.
    ONNX does not support data types smaller than 8-bit integers, whereas in QONNX we are
    interested in smaller integers down to ternary and bipolar."""

    @staticmethod
    def get_accumulator_dt_cands():
        cands = ["BINARY"]
        cands += ["UINT%d" % (x + 1) for x in range(64)]
        cands += ["BIPOLAR", "TERNARY"]
        cands += ["INT%d" % (x + 1) for x in range(64)]
        return cands

    @staticmethod
    def get_smallest_possible(value):
        """Returns smallest (fewest bits) possible DataType that can represent
        value. Prefers unsigned integers where possible."""
        if not int(value) == value:
            return DataType["FLOAT32"]
        cands = DataType.get_accumulator_dt_cands()
        for cand in cands:
            dt = DataType[cand]
            if (dt.min() <= value) and (value <= dt.max()):
                return dt
        raise Exception("Could not find a suitable int datatype for " + str(value))



# Copyright (c) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import onnx
import onnx.helper as oh
import onnx.numpy_helper as nph
from onnx import TensorProto
from onnx.checker import check_model
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.datatype import DataType
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.subpixel_to_deconv import SubPixelToDeconvolution
from qonnx.util.basic import gen_finn_dt_tensor, qonnx_make_model

np.random.seed(0)


def test_subpixel_to_deconv_float_espcn():
    raw_m = get_data("qonnx.data", "onnx/bsd300x3-espcn/subpixel/float_model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    iname = model.graph.input[0].name
    oname = model.graph.output[0].name
    ishape = model.get_tensor_shape(iname)
    rand_inp = gen_finn_dt_tensor(DataType["FLOAT32"], ishape)
    input_dict = {iname: rand_inp}
    expected = oxe.execute_onnx(model, input_dict)[oname]
    new_model = model.transform(SubPixelToDeconvolution())
    # check that there are no DepthToSpace ops left
    op_types = list(map(lambda x: x.op_type, new_model.graph.node))
    assert "DepthToSpace" not in op_types, "Error: the DepthToSpace nodes would be removed."
    produced = oxe.execute_onnx(new_model, input_dict)[oname]
    assert np.isclose(expected, produced, atol=1e-4).all(), "Error: expected output does not match the produced output."


def test_subpixel_to_deconv_quant_espcn():
    # get raw quantized model with reference input
    raw_i = get_data("qonnx.data", "onnx/bsd300x3-espcn/test_data/input_0.pb")
    raw_m = get_data("qonnx.data", "onnx/bsd300x3-espcn/subpixel/quant_model.onnx")
    # create model from the onnx file and infer the shapes
    model = ModelWrapper(raw_m)
    model = model.transform(InferShapes())
    iname = model.graph.input[0].name
    oname = model.graph.output[0].name
    ishape = model.get_tensor_shape(iname)
    # load the reference input tensor
    input_tensor = onnx.load_tensor_from_string(raw_i)
    input_tensor = nph.to_array(input_tensor)
    assert list(input_tensor.shape) == ishape, "Error: reference input doesn't match loaded model."
    input_dict = {iname: input_tensor}
    # get the output from the sub-pixel convolution model
    output_subpixel_conv = oxe.execute_onnx(model, input_dict)[oname]
    # translate the sub-pixel convolution to the deconvolution
    new_model = model.transform(SubPixelToDeconvolution())
    # check that there are no DepthToSpace ops left
    op_types = list(map(lambda x: x.op_type, new_model.graph.node))
    assert "DepthToSpace" not in op_types, "Error: the DepthToSpace nodes would be removed."
    # get the output from the deconvolution model
    output_deconv = oxe.execute_onnx(new_model, input_dict)[oname]
    assert np.isclose(
        output_deconv, output_subpixel_conv, atol=1 / 255.0, rtol=1 / 255.0
    ).all(), "Error: expected output does not match the produced output."


def create_subpixel_conv_model(
    in_channels: int, out_channels: int, input_dim: int, kernel_size: int, upscale_factor: int, bias: bool = False
):
    assert isinstance(kernel_size, int), "Assuming square kernels, so kernel_size needs to be an int."
    padding = (kernel_size - 1) // 2

    ifm_ch = in_channels
    ifm_dim = input_dim
    ofm_ch = out_channels
    ofm_dim = upscale_factor * ifm_dim

    cnv = oh.make_node(
        op_type="Conv",
        inputs=["inp", "W"] if not bias else ["inp", "W", "B"],
        outputs=["hid"],
        kernel_shape=[kernel_size, kernel_size],
        pads=[padding, padding, padding, padding],
        strides=[1, 1],
        group=1,
        dilations=[1, 1],
    )
    d2s = oh.make_node(
        op_type="DepthToSpace",
        inputs=["hid"],
        outputs=["out"],
        blocksize=upscale_factor,
    )

    input_shape = [1, ifm_ch, ifm_dim, ifm_dim]
    output_shape = [1, ofm_ch, ofm_dim, ofm_dim]
    conv_param_shape = [ofm_ch * pow(upscale_factor, 2), ifm_ch, kernel_size, kernel_size]
    bias_param_shape = [ofm_ch * pow(upscale_factor, 2)]

    inp = oh.make_tensor_value_info("inp", TensorProto.FLOAT, input_shape)
    out = oh.make_tensor_value_info("out", TensorProto.FLOAT, output_shape)
    W_conv = oh.make_tensor_value_info("W", TensorProto.FLOAT, conv_param_shape)
    B_conv = oh.make_tensor_value_info("B", TensorProto.FLOAT, bias_param_shape)

    value_info = [W_conv] if not bias else [W_conv, B_conv]

    graph = oh.make_graph(
        nodes=[cnv, d2s],
        name="cnv_graph",
        inputs=[inp],
        outputs=[out],
        value_info=value_info,
    )
    modelproto = qonnx_make_model(graph, producer_name="test_model")
    model = ModelWrapper(modelproto)
    model = model.transform(InferShapes())
    model.set_initializer("W", np.random.rand(*conv_param_shape).astype(np.float32))
    if bias:
        model.set_initializer("B", np.random.rand(*bias_param_shape).astype(np.float32))
    check_model(model._model_proto)
    return model


@pytest.mark.parametrize("kernel_size", [1, 3, 5, 7])
@pytest.mark.parametrize("upscale_factor", [1, 2, 3, 4])
def test_subpixel_to_deconv_layer(kernel_size: int, upscale_factor: int):
    # Create subpixel convolution layer that upsamples a 4x4 image with 1 I/O channel
    model_1 = create_subpixel_conv_model(1, 1, 4, kernel_size, upscale_factor)
    model_2 = model_1.transform(SubPixelToDeconvolution())
    input_shape = [1, 1, 4, 4]
    inp_dict = {"inp": np.random.rand(*input_shape).astype(np.float32)}
    assert oxe.compare_execution(model_1, model_2, inp_dict)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import copy
import numpy as np
import onnx.helper as helper
import onnxruntime as rt
import warnings

import qonnx.analysis.topology as ta
import qonnx.core.execute_custom_node as ex_cu_node
from qonnx.util.basic import (
    get_preferred_onnx_opset,
    get_sanitize_quant_tensors,
    is_finn_op,
    qonnx_make_model,
    sanitize_quant_values,
)


def execute_node(node, context, graph, return_full_exec_context=False, opset_version=get_preferred_onnx_opset()):
    """Executes a single node by using onnxruntime or with a custom function.

    Input/output provided via context."""

    if is_finn_op(node.domain):
        ex_cu_node.execute_custom_node(node, context, graph, onnx_opset_version=opset_version)
    else:
        # onnxruntime unfortunately does not implement run_node as defined by ONNX,
        # it can only execute entire models -- so we create a model which solely
        # consists of our current node.
        # note: ensure that the same ValueInfo does not appear both in
        # graph.value_info as well as graph.output or graph.input
        # nodes with multiple outputs that are a mix of value_info and
        # input/outputs may get them reordered below
        # note: a node's input may (also) be a top-level input or output
        node_inputs = list(filter(lambda x: x.name in node.input, graph.input))
        node_inputs += list(filter(lambda x: x.name in node.input, graph.output))
        node_inputs += list(filter(lambda x: x.name in node.input, graph.value_info))
        node_outputs = list(filter(lambda x: x.name in node.output, graph.output))
        node_outputs += list(filter(lambda x: x.name in node.output, graph.value_info))
        for attr in node.attribute:
            if attr.type == 5:
                subgraph = attr.g
                for subgraph_node in subgraph.node:
                    subgraph_node_inputs = list(filter(lambda x: x.name in subgraph_node.input, graph.value_info))
                    new_inps = list(filter(lambda x: x not in node_inputs, subgraph_node_inputs))
                    node_inputs += new_inps
        node_graph = helper.make_graph(
            nodes=[node],
            name="single-node-exec",
            inputs=node_inputs,
            outputs=node_outputs,
        )
        node_model = qonnx_make_model(node_graph)
        node_model.opset_import[0].version = opset_version
        input_dict = dict()
        for inp in node.input:
            input_dict[inp] = context[inp]

        sess = rt.InferenceSession(node_model.SerializeToString())
        output_list = sess.run(None, input_dict)

        for output_ind in range(len(node.output)):
            # get the name of the target buffer from node.output
            outp = node.output[output_ind]

            # retrieve the index of that name in node_outputs
            for i in range(len(node_outputs)):
                if outp == node_outputs[i].name:
                    list_ind = i

            # use that index to index output_list
            if output_list[list_ind].shape != context[outp].shape:
                warnings.warn(
                    """Output shapes disagree after node %s execution:
                    found %s vs expected %s"""
                    % (str(node), str(output_list[list_ind].shape), str(context[outp].shape))
                )
            context[outp] = output_list[list_ind]


def execute_onnx(model, input_dict, return_full_exec_context=False, start_node=None, end_node=None):
    """Executes given ONNX ModelWrapper with given named inputs.

    If return_full_exec_context is False, a dict of named outputs is returned
    as indicated by the model.graph.output.

    If return return_full_exec_context is True, the full set of tensors used by
    the execution (including inputs, weights, activations and final outputs)
    will be returned as a dict.

    When start_node and end_node are set to None, the whole graph is executed.
    If they are set to particular ONNX nodes, only the subgraph between (and
    including) those nodes is executed.
    """

    if not model.check_all_tensor_shapes_specified():
        raise Exception("Found unspecified tensor shapes, try infer_shapes")
    ret = model.analysis(ta.nodes_topologically_sorted)
    assert (
        ret["nodes_topologically_sorted"] is True
    ), """Nodes must be
    topologically sorted."""

    graph = model.graph
    # first, we need to make sure that every variable required by the graph has
    # some buffer associated with it. this includes graph inputs (which includes
    # the input data as well as the trained parameters) and the graph ValueInfo
    # (intermediate tensors between layers)
    # this is provided by the execution_context, which is a dict of np.ndarray
    execution_context = model.make_empty_exec_context()
    # fill in any inputs provided to this function
    for inp_name in input_dict.keys():
        if inp_name in execution_context:
            if execution_context[inp_name].shape == input_dict[inp_name].shape:
                execution_context[inp_name] = input_dict[inp_name]
            else:
                raise Exception(
                    "Shape mismatch for provided input %s: found %s expected %s "
                    % (
                        inp_name,
                        str(execution_context[inp_name].shape),
                        str(input_dict[inp_name].shape),
                    )
                )
        # else:
        # raise Exception("Provided input not found in graph context: %s" % inp_name)

    # check if model has an execution mode set
    # if None, execute model node by node using execute_node()
    model_exec_mode = model.get_metadata_prop("exec_mode")
    if (model_exec_mode is None) or (model_exec_mode == ""):
        # extract opset version for node-by-node execution
        opset_version = model.model.opset_import[0].version
        # execute the model node by node
        # we can simply walk down the list since the ONNX spec guarantees that it is
        # topologically sorted
        subgraph = []
        if start_node is None:
            start_node = model.graph.node[0]
        if end_node is None:
            end_node = model.graph.node[-1]
        # select the nodes between specified start/end nodes
        start_ind = model.get_node_index(start_node)
        end_ind = model.get_node_index(end_node) + 1
        assert end_ind >= start_ind, "Start/end nodes must define valid subgraph"
        subgraph = graph.node[start_ind:end_ind]
        for node in subgraph:
            if get_sanitize_quant_tensors() != 0:
                # round input values to match quantization annotation
                execution_context = sanitize_quant_values(model, node.input, execution_context)
            execute_node(node, execution_context, graph, return_full_exec_context, opset_version)
            if get_sanitize_quant_tensors() != 0:
                # round output values to quantization annotation
                execution_context = sanitize_quant_values(model, node.output, execution_context)
    else:
        raise Exception('Metadata property "exec_mode" is set to an unknown value.')

    if return_full_exec_context:
        return execution_context
    else:
        # provide outputs as dict
        output_dict = dict()
        for out_tensor in graph.output:
            out_name = out_tensor.name
            output_dict[out_name] = execution_context[out_name]
        return output_dict


def execute_onnx_and_make_model(model, input_dict):
    """Executes given ONNX ModelWrapper with given named inputs and return a new
    ModelWrapper where an initializer is provided for each tensor as taken from
    the execution. This new model is useful for debugging, since it contains
    all the intermediate activation values."""

    # retrieve the full execution context
    execution_context = execute_onnx(model, input_dict, True)
    new_model = copy.deepcopy(model)
    # create value_info entries and initializers for everything
    for i in execution_context.keys():
        new_model.set_initializer(i, execution_context[i])
    for vi in new_model.graph.value_info:
        new_model.graph.output.append(vi)
    return new_model


def compare_execution(
    model_a,
    model_b,
    input_dict,
    compare_fxn=lambda x, y: np.isclose(x, y, atol=1e-3).all(),
):
    """Executes two ONNX models and compare their outputs using given function.

    compare_fxn should take in two tensors and return a Boolean"""
    # compare values from first output tensors produced
    res_a = list(execute_onnx(model_a, input_dict).items())[0][1]
    res_b = list(execute_onnx(model_b, input_dict).items())[0][1]
    return compare_fxn(res_a, res_b)



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import copy
import pathlib
import tempfile
from onnx import helper

from qonnx.transformation.base import Transformation


class PartitionFromLambda(Transformation):
    """Split a graph into partitions. Each resulting partition node has a model
    attribute indicating the path to the subordinate onnx file.
    Cleanup and InferShapes() transformations should be applied first.

    Argument 0: partitioning
    * Function performing the mapping: node -> partition_id (int or string)
    * Partitions may not cover the graph completely (nodes mapped to -1 are retained)
    * Mapping must return -1 for GenericPartition nodes

    Argument 1 (optional): partition_dir
    * Manually define where to save the partition models
    """

    def __init__(self, partitioning=lambda node: -1, partition_dir=None):
        super().__init__()
        self.partitioning = partitioning
        self.partition_dir = partition_dir

    def apply(self, model):
        # identify partitions to create
        original_nodes = list(model.graph.node)
        partition_ids = set(list(map(self.partitioning, original_nodes)))
        partition_ids.discard(-1)

        # prepare dir for generated .onnx models
        if self.partition_dir is None:
            self.partition_dir = tempfile.mkdtemp(prefix="partitioning_")
        else:
            pathlib.Path(self.partition_dir).mkdir(parents=True, exist_ok=True)

        for partition_id in partition_ids:
            all_nodes = list(model.graph.node)
            partition_nodes = list(filter(lambda x: self.partitioning(x) == partition_id, all_nodes))
            non_partition_nodes = list(filter(lambda x: x not in partition_nodes, all_nodes))

            # partition the model into two models
            p_model = copy.deepcopy(model)
            non_p_model = model
            # remove all non-partition nodes from the partition model
            for node_to_remove in non_partition_nodes:
                p_model.graph.node.remove(node_to_remove)

            # identify the entry and exit points for the partition part
            p_in = []
            p_out = []
            p_start_ind = 0
            for node in p_model.graph.node:
                for in_tensor in node.input:
                    # check if producer has been removed = lies outside the partition
                    has_initializer = in_tensor in [x.name for x in p_model.graph.initializer]
                    has_producer = p_model.find_producer(in_tensor) is not None
                    if not has_initializer and not has_producer:
                        # the same tensor could feed multiple nodes within the partition
                        # (e.g. for residual connections), so we avoid duplicates
                        if in_tensor not in p_in:
                            p_in.append(in_tensor)
                        # keep track of where this partition starts topologically
                        if p_start_ind == 0:
                            p_start_ind = all_nodes.index(node)
                for out_tensor in node.output:
                    # check if tensor is top-level output
                    # or has a consumer outside the partition
                    if out_tensor in [x.name for x in model.graph.output]:
                        if out_tensor not in p_out:
                            p_out.append(out_tensor)
                    else:
                        for consumer in model.find_consumers(out_tensor):
                            if self.partitioning(consumer) != partition_id:
                                if out_tensor not in p_out:
                                    p_out.append(out_tensor)

            p_in_vi = list(map(lambda x: p_model.get_tensor_valueinfo(x), p_in))
            p_out_vi = list(map(lambda x: p_model.get_tensor_valueinfo(x), p_out))

            # check if partitioning is legal (i.e. creates no cycles)
            to_check = [model.find_producer(x) for x in p_in]
            while len(to_check) > 0:
                next_to_check = []
                for node in to_check:
                    if node is not None:
                        assert (
                            self.partitioning(node) != partition_id
                        ), """cycle-free graph violated: partition depends on itself"""
                        # print(node)
                        predecessors = model.find_direct_predecessors(node)
                        if predecessors is not None:
                            next_to_check.extend(predecessors)
                to_check = next_to_check

            # set p graph in/out to be p_in/p_out
            while len(p_model.graph.input) > 0:
                p_model.graph.input.pop()
            for i in p_in_vi:
                p_model.graph.input.append(i)

            while len(p_model.graph.output) > 0:
                p_model.graph.output.pop()
            for o in p_out_vi:
                p_model.graph.output.append(o)

            # remove redundant input and output value_info entries
            for i in p_in_vi:
                if i in p_model.graph.value_info:
                    p_model.graph.value_info.remove(i)

            for o in p_out_vi:
                if o in p_model.graph.value_info:
                    p_model.graph.value_info.remove(o)

            # save partition model
            p_model_filename = self.partition_dir + "/partition_" + str(partition_id) + ".onnx"
            p_model.cleanup()
            p_model.save(p_model_filename)

            # insert GenericPartition node
            p_node = helper.make_node(
                "GenericPartition",
                p_in,
                p_out,
                name="GenericPartition_" + str(partition_id),
                # use the model attribute to mark the partition model
                model=p_model_filename,
                domain="qonnx.custom_op.general",
            )
            non_p_model.graph.node.insert(p_start_ind, p_node)

            # remove all partition nodes from the parent model
            # do this after inserting the p_node for easier p_start_ind handling
            for node_to_remove in partition_nodes:
                non_p_model.graph.node.remove(node_to_remove)

            model = non_p_model

        return (model, False)


class PartitionFromDict(Transformation):
    """Split a graph into partitions. Each resulting partition node has a model
    attribute indicating the path to the subordinate onnx file.
    Cleanup and InferShapes() transformations should be applied first.

    This transformation builds on PartitionFromLambda() and takes a dictionary that
    defines partitions based on node indices.

    Argument 0: partitioning
    * Dictionary with the following format:  { partition_id : node_index_list }
    * Example: {0 : [3,4,5], 1 : range(10, 15)}

    Argument 1 (optional): partition_dir
    * Manually define where to save the partition models
    """

    def __init__(self, partitioning={}, partition_dir=None):
        super().__init__()
        self.partitioning = partitioning
        self.partition_dir = partition_dir

    def apply(self, model):
        # prepare node -> int assignment fct.
        def partitioning_func(node):
            if node not in model.graph.node:
                return -1
            node_index = list(model.graph.node).index(node)
            candidates = list(filter(lambda key_value: node_index in key_value[1], self.partitioning.items()))
            if len(candidates) == 0:
                return -1
            assert len(candidates) == 1, f"single node assigned to multiple partitions: {candidates}"
            return candidates[0][0]  # partition_id

        # apply partitioning
        model = model.transform(PartitionFromLambda(partitioning_func, self.partition_dir))
        return (model, False)



# Copyright (c) 2022 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np

import qonnx.util.basic as util
from qonnx.core.datatype import DataType

"""Calculate the minimum and maximum possible result (accumulator) values
for a dot product x * A, given matrix A of dims (MW, MH), and vector (1, MW)
with datatype vec_dt. Returns (acc_min, acc_max).
"""

np.random.seed(0)

datatypes = [
    DataType["UINT8"],
    DataType["INT8"],
    DataType["UINT3"],
    DataType["INT3"],
    DataType["BIPOLAR"],
    DataType["TERNARY"],
]


@pytest.mark.parametrize("wdt", datatypes)
@pytest.mark.parametrize("idt", datatypes)
@pytest.mark.parametrize("size", [1, 10, 100])
def test_calculate_matvec_accumulator_range(wdt: DataType, idt: DataType, size: int):
    weights = util.gen_finn_dt_tensor(wdt, tensor_shape=(size, 1))  # (MW, MH)
    acc_min, acc_max = util.calculate_matvec_accumulator_range(weights, idt)
    for _ in range(1000):
        inputs = util.gen_finn_dt_tensor(idt, tensor_shape=(1, size))
        acc_vals = inputs @ weights
        assert acc_vals.min() >= acc_min
        assert acc_vals.max() <= acc_max



# Copyright (c) 2024, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import warnings
from onnx import helper

from qonnx.core.datatype import DataType
from qonnx.custom_op.general.quant import quant, resolve_rounding_mode
from qonnx.transformation.base import Transformation
from qonnx.util.basic import auto_pad_to_explicit_padding, get_by_name


def _weight_convolution(cnv_weights: np.ndarray, scale: int) -> np.ndarray:
    """Adaptation of the weight convolution algorithm as proposed in Colbert et al. (2021) - `An
    Energy-Efficient Edge Computing Paradigm for Convolution-Based Image Upsampling`"""
    ofm_ch = cnv_weights.shape[0]
    ifm_ch = cnv_weights.shape[1]
    kh_size = cnv_weights.shape[2]
    kw_size = cnv_weights.shape[3]
    assert kh_size == kw_size, "Only square channels supported currently."
    # NOTE - this is different than the convolution kernels, which are OC x IC x KH x KW
    #        rather than IC x OC x KH x KW
    dcnv_weights = np.zeros((ifm_ch, ofm_ch, kh_size + scale - 1, kw_size + scale - 1))
    for oc in range(ofm_ch):
        for ic in range(ifm_ch):
            for i in range(scale):
                for j in range(scale):
                    dcnv_weights[ic, oc, i : i + kh_size, j : j + kw_size] += np.rot90(cnv_weights[oc, ic], 2, [0, 1])
    return dcnv_weights


class ResizeConvolutionToDeconvolution(Transformation):
    """Replaces resize convolution layers (e.g., nearest neighbor upsample + same-padded convolution)
    with deconvolution layers using the weight convolution algorithm. Currently does not support
    resize convolutions that use bilinear or bicubic upsampling"""

    def __init__(self, maintain_bit_width: bool = False):
        super().__init__()
        self.maintain_bit_width = maintain_bit_width

    def apply(self, model):
        graph = model.graph
        node_ind = 0
        graph_modified = False
        for n in graph.node:
            node_ind += 1
            if n.op_type == "Resize":
                resize_input = n.input[0]
                resize_output = n.output[0]
                consumers = model.find_consumers(resize_output)

                if len(consumers) == 0:
                    continue

                if len(consumers) > 1 and any([c.op_type == "Conv" for c in consumers]):
                    warnings.warn("Skipping resize conv that has resize with multiple consumers. Not yet supported.")
                    continue

                conv = consumers[0]
                if conv is not None and conv.op_type == "Conv":
                    # TODO: extend support to other resize convolutions
                    resize_mode = get_by_name(n.attribute, "mode").s.decode()
                    if resize_mode != "nearest":
                        warnings.warn(f"Skipping resize conv with resize_mode={resize_mode}. Not yet supported.")
                        continue

                    group = get_by_name(conv.attribute, "group").i
                    if group != 1:
                        warnings.warn("Skipping resize conv with group > 1. Not yet supported.")
                        continue

                    # The weights of the convolution can be generated by another input op if the model is
                    # quantized. Preliminary support for quantization focuses on QONNX ops (i.e., Quant)
                    weight_name = conv.input[1]
                    weight_prod = model.find_producer(weight_name)

                    # If the producer is None, then it is initialized by the Conv node
                    if weight_prod is None:
                        W_conv = model.get_initializer(weight_name)  # (OC, IC, KH, KW)

                    # If the convolution weights are not initialized by the convolution, then we need to
                    # find the node is producing the weights
                    else:
                        if weight_prod.op_type == "Quant":
                            [q_w_name, q_s_name, q_zp_name, q_bw_name] = weight_prod.input
                            W_conv = model.get_initializer(q_w_name)
                            W_scale = model.get_initializer(q_s_name)
                            if isinstance(W_scale, np.ndarray) and W_scale.ndim > 1:
                                W_scale = np.moveaxis(W_scale, 0, 1)
                            W_zeropt = model.get_initializer(q_zp_name)
                            if isinstance(W_zeropt, np.ndarray) and W_zeropt.ndim > 1:
                                W_zeropt = np.moveaxis(W_zeropt, 0, 1)
                            W_bitwidth = model.get_initializer(q_bw_name)
                            W_signed = get_by_name(weight_prod.attribute, "signed").i
                            W_narrow = get_by_name(weight_prod.attribute, "narrow").i
                            W_rounding_mode = get_by_name(weight_prod.attribute, "rounding_mode").s.decode()
                        else:
                            warnings.warn(
                                f"Weight producer is {weight_prod.op_type}, not a QONNX Quant node. Not yet supported."
                            )
                            continue

                    kshape = get_by_name(conv.attribute, "kernel_shape").ints
                    idim = model.get_tensor_shape(conv.input[0])  # require NCHW
                    odim = model.get_tensor_shape(conv.output[0])  # require NCHW
                    if not (len(odim) == len(idim) == 4):
                        warnings.warn("Skipping resize conv, only 2D convolutions supported.")
                        continue

                    [_, ifm_ch, ifm_dim_h, ifm_dim_w] = idim
                    [_, ofm_ch, ofm_dim_h, ofm_dim_w] = odim

                    if (ifm_dim_h != ofm_dim_h) or (ifm_dim_w != ofm_dim_w):
                        warnings.warn("Skipping resize conv, only same-padded convs supported.")
                        continue
                    dilation_attr = get_by_name(conv.attribute, "dilations")
                    if dilation_attr is not None:
                        dilation = dilation_attr.ints
                    else:
                        dilation = [1, 1]  # default value
                    if dilation != [1, 1]:
                        warnings.warn("Skipping resize conv, only supporting dilation=[1,1].")
                        continue
                    # get resize scaling attribute
                    resize_scales = model.get_initializer(n.input[2])  # assume NCHW
                    if not (resize_scales[0] == resize_scales[1] == 1):
                        warnings.warn("Skipping resize conv, scaling along batch or channel dimension not supported.")
                        continue
                    if resize_scales[2] != resize_scales[3]:
                        warnings.warn("Skipping resize conv, non-square scaling not yet supported.")
                        continue
                    resize_scale = int(resize_scales[2])  # TODO: extend to vector once non-square scaling supported

                    W_deconv = _weight_convolution(W_conv, resize_scale).astype(np.float32)
                    kh_size_deconv = kshape[0] + resize_scale - 1
                    kw_size_deconv = kshape[1] + resize_scale - 1
                    assert W_deconv.shape == (
                        ifm_ch,
                        ofm_ch,
                        kh_size_deconv,
                        kw_size_deconv,
                    ), "The resulting deconvolution weight shape is incorrect."

                    stride_h = get_by_name(conv.attribute, "strides").ints[0]
                    stride_w = get_by_name(conv.attribute, "strides").ints[1]
                    # handle both auto_pad and explicit padding
                    auto_pad = get_by_name(conv.attribute, "auto_pad")
                    if auto_pad is not None:
                        # find equivalent specified padding
                        auto_pad = auto_pad.s.decode("utf-8")
                        if auto_pad == "NOTSET":
                            # use specified padding
                            pad = get_by_name(conv.attribute, "pads").ints
                        else:
                            pad = auto_pad_to_explicit_padding(
                                auto_pad,
                                ifm_dim_h,
                                ifm_dim_w,
                                kshape[0],
                                kshape[1],
                                stride_h,
                                stride_w,
                                len(model.get_tensor_shape(n.input[0])) - 2,
                            )
                    else:
                        # use specified padding
                        pad = get_by_name(conv.attribute, "pads").ints

                    # if `maintain_bit_width`, then we use the quant parameters to
                    # re-quantize the weights after the weight convolution
                    if self.maintain_bit_width and (weight_prod is not None):
                        W_deconv_quant = quant(W_deconv, W_scale, W_zeropt, W_bitwidth, W_signed, W_narrow, W_rounding_mode)
                        if not np.allclose(W_deconv, W_deconv_quant):
                            warnings.warn("Clipping error introduced, consider `maintain_bit_width=False`.")

                    # if not `maintain_bit_width`, then we adjust the bit width to
                    # account for the clipping errors.
                    elif weight_prod is not None:
                        round_fnc = resolve_rounding_mode(W_rounding_mode)
                        W_int = (W_deconv / W_scale) + W_zeropt
                        W_int = round_fnc(W_int)  # handling rounding errors
                        W_min = W_int.min()
                        W_max = W_int.max()
                        if W_min < 0:
                            if abs(W_min) > W_max:
                                wdt = DataType.get_smallest_possible(W_min)
                            else:
                                wdt = DataType.get_smallest_possible(-W_max - 1)
                        else:
                            wdt = DataType.get_smallest_possible(W_max)
                        assert np.vectorize(wdt.allowed)(W_int).all(), "Error: issue finding data type to support."
                        if W_bitwidth != wdt.bitwidth():
                            W_bitwidth = np.array(wdt.bitwidth(), dtype=np.float32)
                        assert wdt.signed() == W_signed, "Error: should maintain sign of the weights."

                    deconv_inps = [resize_input, weight_name]
                    # Make sure to keep the biases from the convolution
                    if len(conv.input) == 3:
                        bias_name = conv.input[2]
                        bias_prod = model.find_producer(bias_name)
                        # If the producer is None, then it is initialized by the Conv node
                        # and we need to ensure it isn't removed with the Conv node
                        if bias_prod is None:
                            B_conv = model.get_initializer(bias_name)  # (OC,)
                            model.set_initializer(bias_name, B_conv)
                        deconv_inps.append(bias_name)  # add to the inputs
                    deconv_outs = conv.output
                    deconv_pad = pad
                    deconv_node = helper.make_node(
                        "ConvTranspose",
                        deconv_inps,
                        deconv_outs,
                        kernel_shape=[kh_size_deconv, kw_size_deconv],
                        strides=[resize_scale, resize_scale],
                        pads=deconv_pad,
                        group=group,
                        dilations=dilation,
                    )
                    W_deconv_init = weight_name
                    if weight_prod is not None:
                        W_deconv_init = q_w_name
                        model.set_initializer(q_zp_name, W_zeropt)
                        model.set_initializer(q_s_name, W_scale)
                        model.set_initializer(q_bw_name, W_bitwidth)
                    model.set_initializer(W_deconv_init, W_deconv)
                    model.set_tensor_shape(weight_name, list(W_deconv.shape))
                    graph.node.insert(node_ind, deconv_node)
                    # remove old nodes
                    graph.node.remove(n)
                    graph.node.remove(conv)
                    graph_modified = True

        return (model, graph_modified)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import os

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.qcdq_to_qonnx import QCDQToQuant
from qonnx.transformation.qonnx_to_qcdq import QuantToQCDQ
from qonnx.util.cleanup import cleanup_model
from qonnx.util.test import download_model, get_golden_in_and_output, test_model_details

qonnxtoqcdq_details = {
    "FINN-CNV_W2A2": {
        "nonconvertible_quant": 0,
        "exp_qdq_nodes": 18,
        # input quantizer doesn't need Clip so 1 less
        "exp_clip_nodes": 17,
    },
    "FINN-TFC_W2A2": {
        # all Quant nodes convertible to QCDQ
        "nonconvertible_quant": 0,
        "exp_qdq_nodes": 8,
        "exp_clip_nodes": 8,
    },
    "RadioML_VGG10": {
        # 23 bit bias quant not convertible to QCDQ
        "nonconvertible_quant": 1,
        "exp_qdq_nodes": 20,
        # half the Quants don't need Clip (not signed narrow)
        "exp_clip_nodes": 10,
    },
    "MobileNetv1-w4a4": {
        # 18 bit bias quant not convertible to QCDQ
        "nonconvertible_quant": 1,
        "exp_qdq_nodes": 55,
        "exp_clip_nodes": 55,
    },
}

# inherit basics for matching testcases from test util
model_details = {k: v for (k, v) in test_model_details.items() if k in qonnxtoqcdq_details.keys()}
model_details = {**model_details, **qonnxtoqcdq_details}


@pytest.mark.parametrize("test_model", model_details.keys())
def test_qonnx_to_qcdq_to_qonnx(test_model):
    test_details = model_details[test_model]
    dl_file = download_model(test_model=test_model)
    assert os.path.isfile(dl_file)
    model = ModelWrapper(dl_file)
    model = cleanup_model(model)
    input_tensor, golden_result = get_golden_in_and_output(test_model)
    # test Quant -> QCDQ conversion
    model = model.transform(QuantToQCDQ())
    assert len(model.get_nodes_by_op_type("Quant")) == test_details["nonconvertible_quant"]
    assert len(model.get_nodes_by_op_type("QuantizeLinear")) > 0
    assert len(model.get_nodes_by_op_type("DequantizeLinear")) > 0
    assert len(model.get_nodes_by_op_type("Clip")) == test_details["exp_clip_nodes"]
    model = cleanup_model(model)
    input_dict = {model.graph.input[0].name: input_tensor}
    produced_output_dict = oxe.execute_onnx(model, input_dict)
    produced_result = produced_output_dict[model.graph.output[0].name]
    assert np.isclose(golden_result, produced_result).all()
    # now test QCDQ -> Quant conversion
    model = model.transform(QCDQToQuant())
    assert len(model.get_nodes_by_op_type("QuantizeLinear")) == 0
    assert len(model.get_nodes_by_op_type("DequantizeLinear")) == 0
    assert len(model.get_nodes_by_op_type("Clip")) == 0
    model = cleanup_model(model)
    new_output_dict = oxe.execute_onnx(model, input_dict)
    roundtrip_result = new_output_dict[model.graph.output[0].name]
    assert np.isclose(golden_result, roundtrip_result).all()
    os.unlink(dl_file)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
from onnx import TensorProto, helper

import qonnx.analysis.topology as ta
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.general import SortGraph
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.basic import qonnx_make_model


def make_randomly_sorted_linear_model(num_of_nodes, seed=None):
    if seed is not None:
        np.random.seed(seed)

    ch = 2
    ifmdim = 16
    input_shape = (1, ch, ifmdim, ifmdim)

    top_in = helper.make_tensor_value_info("t0", TensorProto.FLOAT, input_shape)
    top_out = helper.make_tensor_value_info("t" + str(num_of_nodes), TensorProto.FLOAT, input_shape)

    value_info = []
    nodes = []
    for i in range(num_of_nodes):
        nodes += [helper.make_node("Add", ["t" + str(i), "p" + str(i)], ["t" + str(i + 1)])]
        value_info += [helper.make_tensor_value_info("p" + str(i), TensorProto.FLOAT, input_shape)]

    nodes = np.random.permutation(nodes)

    modelproto = qonnx_make_model(
        helper.make_graph(
            name="test",
            inputs=[top_in],
            outputs=[top_out],
            value_info=value_info,
            nodes=nodes,
        )
    )
    model = ModelWrapper(modelproto)
    model = model.transform(InferShapes())

    for i in range(num_of_nodes):
        model.set_initializer("p" + str(i), np.random.rand(*input_shape).astype(np.float32))

    return model


@pytest.mark.parametrize("num_of_nodes", [64])
def test_sort_linear_graph(num_of_nodes):
    model = make_randomly_sorted_linear_model(num_of_nodes, seed=0)
    new_model = model.transform(SortGraph())

    # Test
    ret = new_model.analysis(ta.nodes_topologically_sorted)
    assert ret["nodes_topologically_sorted"], "Nodes are not topologically sorted."


def test_sort_nonlinear_graph():
    ch = 2
    ifmdim = 16
    input_shape = (1, ch, ifmdim, ifmdim)

    top_in = helper.make_tensor_value_info("top_in", TensorProto.FLOAT, input_shape)
    top_out = helper.make_tensor_value_info("top_out", TensorProto.FLOAT, input_shape)

    num_of_params = 8
    value_info = []
    for i in range(num_of_params):
        value_info += [helper.make_tensor_value_info("p" + str(i), TensorProto.FLOAT, input_shape)]

    modelproto = qonnx_make_model(
        helper.make_graph(
            name="test",
            inputs=[top_in],
            outputs=[top_out],
            value_info=value_info,
            nodes=[
                # Not sorted nodes
                helper.make_node("Mul", ["fork1", "p2"], ["t3"]),
                helper.make_node("Add", ["t4", "p3"], ["t5"]),
                helper.make_node("Add", ["t2", "t3"], ["t4"]),
                helper.make_node("Add", ["t6", "t7"], ["t8"]),
                helper.make_node("Add", ["fork3", "fork3"], ["top_out"]),
                helper.make_node("Mul", ["t5", "p4"], ["fork2"]),
                helper.make_node("Add", ["top_in", "p0"], ["fork1"]),
                helper.make_node("Mul", ["fork1", "p1"], ["t2"]),
                helper.make_node("Add", ["fork2", "p5"], ["t6"]),
                helper.make_node("Add", ["fork2", "p6"], ["t7"]),
                helper.make_node("Mul", ["t8", "p7"], ["fork3"]),
            ],
        )
    )
    model = ModelWrapper(modelproto)
    model = model.transform(InferShapes())

    np.random.seed(0)
    for i in range(num_of_params):
        model.set_initializer("p" + str(i), np.random.rand(*input_shape).astype(np.float32))

    new_model = model.transform(SortGraph())

    # Test
    ret = new_model.analysis(ta.nodes_topologically_sorted)
    assert ret["nodes_topologically_sorted"], "Nodes are not topologically sorted."


if __name__ == "__main__":
    import time

    sizes = [10, 50, 100, 500, 1000]
    times = []
    reps = 10

    print("SortGraph performance test:")
    print("Test sizes", sizes)
    print("Repetitions per size:", reps)
    for sz in sizes:
        acc_time = 0
        print(" Testing size ", sz)
        for i in range(reps):
            # it should take the same time even with the sorted one
            # but better new model each time as it is a more general approach
            model = make_randomly_sorted_linear_model(sz)  # new model as seed is None
            bef = time.time()
            new_model = model.transform(SortGraph(), make_deepcopy=False)
            acc_time += time.time() - bef

        times += [acc_time / reps]

    # print csv
    print("\nnum_of_nodes,  seconds")
    for sz, tm in zip(sizes, times):
        print("{:12d}, {:6.4e}".format(sz, tm))

    # plot
    # import matplotlib.pyplot as plt
    # plt.plot(sizes,times,"--o")
    # plt.grid(True)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import onnx
import onnx.numpy_helper as nph
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.general import GiveReadableTensorNames, GiveUniqueNodeNames
from qonnx.transformation.infer_datatypes import InferDataTypes
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.insert_topk import InsertTopK


@pytest.mark.parametrize("k", [1, 2])
def test_topk_insert(k):
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model.model.opset_import[0].version = 11

    # do transformations (no topk)
    model = model.transform(InferShapes())
    model = model.transform(FoldConstants())
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    model = model.transform(InferDataTypes())

    # verification: generate random input, run through net, streamline,
    # run again, check that output is top-k
    raw_i = get_data("qonnx.data", "onnx/mnist-conv/test_data_set_0/input_0.pb")
    input_tensor = onnx.load_tensor_from_string(raw_i)
    input_tensor = nph.to_array(input_tensor)
    input_dict = {"global_in": input_tensor}
    output_golden = oxe.execute_onnx(model, input_dict)["global_out"]
    output_golden_topk = np.flip(output_golden.flatten().argsort())[:k]
    output_golden_topk = output_golden_topk.flatten()

    # insert top-k
    model = model.transform(InsertTopK(k))
    model = model.transform(GiveUniqueNodeNames())
    model = model.transform(GiveReadableTensorNames())
    model = model.transform(InferShapes())

    # verify output of top-k
    output_dict_topk = oxe.execute_onnx(model, input_dict)
    output_pysim_topk = output_dict_topk[list(output_dict_topk.keys())[0]]
    output_pysim_topk = output_pysim_topk.astype(np.int32).flatten()

    assert np.array_equal(output_golden_topk, output_pysim_topk)



# Copyright (c) 2024 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from pkgutil import get_data

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.channels_last import ConvertToChannelsLastAndClean
from qonnx.util.basic import gen_finn_dt_tensor


def test_channelslast_residual():
    raw_m = get_data("qonnx.data", "onnx/residual_block_clean.onnx")
    model = ModelWrapper(raw_m)
    iname = model.graph.input[0].name
    idt = model.get_tensor_datatype(iname)
    ishape = model.get_tensor_shape(iname)
    idict = {iname: gen_finn_dt_tensor(idt, ishape)}
    oname = model.graph.output[0].name
    expected_out = oxe.execute_onnx(model, idict)[oname]
    model = model.transform(ConvertToChannelsLastAndClean(make_input_channels_last=False))
    expected_ops = ["Transpose", "Conv", "Conv", "Relu", "Conv", "Relu", "Add", "MaxPool", "Transpose"]
    ops = [x.op_type for x in model.graph.node]
    assert ops == expected_ops, "Did not found expected op sequence after lowering and channels-last"
    for node in model.graph.node:
        if node.op_type in ["Conv", "MaxPool"]:
            assert node.domain == "qonnx.custom_op.channels_last"
    out = oxe.execute_onnx(model, idict)[oname]
    assert np.isclose(expected_out, out, atol=1e-4).all()



# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

import torch
from brevitas.export.onnx import onnx_export_opset
from torch.autograd import Function

AXIS_OPSET = 13
DOMAIN_STRING = "onnx.brevitas"


class DequantizeLinearFn(Function):
    @staticmethod
    def symbolic(g, x, input_scale, input_zero_point, input_axis):
        opset_version = onnx_export_opset()

        if input_axis is not None and opset_version < AXIS_OPSET:
            raise RuntimeError("ONNX Opset 13 is required for per-channel quantization")
        elif input_axis is not None and opset_version >= AXIS_OPSET:
            ret = g.op("DequantizeLinear", x, input_scale, input_zero_point, axis_i=input_axis)
        else:
            ret = g.op("DequantizeLinear", x, input_scale, input_zero_point)
        return ret

    @staticmethod
    def forward(ctx, int_x, input_scale, input_zero_point, input_axis):
        return int_x.float()


class IntClipFn(Function):
    @staticmethod
    def symbolic(g, int_x, min_int_val, max_int_val):
        ret = g.op("Clip", int_x, min_int_val, max_int_val)
        return ret

    @staticmethod
    def forward(ctx, int_x, min_int_val, max_int_val):
        return int_x


class QuantizeLinearFn(Function):
    @staticmethod
    def symbolic(g, x, output_scale, ouput_zero_point, output_dtype, output_axis):
        opset_version = onnx_export_opset()

        if output_axis is not None and opset_version < AXIS_OPSET:
            raise RuntimeError("ONNX Opset 13 is required for per-channel quantization")
        elif output_axis is not None and opset_version >= AXIS_OPSET:
            ret = g.op("QuantizeLinear", x, output_scale, ouput_zero_point, axis_i=output_axis)
        else:
            ret = g.op("QuantizeLinear", x, output_scale, ouput_zero_point)
        return ret

    @staticmethod
    def forward(ctx, x, output_scale, ouput_zero_point, output_dtype, output_axis):
        return x.type(output_dtype)


class BrevitasQuantLSTMCellFn(Function):
    @staticmethod
    def symbolic(
        g,  # args and kwargs passed from _QuantLSTMLayer
        quant_input,
        quant_hidden_state,
        quant_cell_state,
        quant_weight_ii,
        quant_weight_if,
        quant_weight_ic,
        quant_weight_io,
        quant_weight_hi,
        quant_weight_hf,
        quant_weight_hc,
        quant_weight_ho,
        quant_bias_input,
        quant_bias_forget,
        quant_bias_cell,
        quant_bias_output,  # Symbolic kwargs passed from BrevitasQuantLSTMLayerHandler
        batch_first,
        reverse_input,
        cifg,  # Output quant
        output_scale,
        output_zero_point,
        output_bit_width,
        output_narrow_range,
        output_signed,
        output_rounding_mode,  # Cell state quant
        cell_state_scale,
        cell_state_zero_point,
        cell_state_bit_width,
        cell_state_narrow_range,
        cell_state_signed,
        cell_state_rounding_mode,  # Input gate accumulator quant
        input_acc_scale,
        input_acc_zero_point,
        input_acc_bit_width,
        input_acc_narrow_range,
        input_acc_signed,
        input_acc_rounding_mode,  # Forget gate accumulator quant
        forget_acc_scale,
        forget_acc_zero_point,
        forget_acc_bit_width,
        forget_acc_narrow_range,
        forget_acc_signed,
        forget_acc_rounding_mode,  # Cell gate accumulator quant
        cell_acc_scale,
        cell_acc_zero_point,
        cell_acc_bit_width,
        cell_acc_narrow_range,
        cell_acc_signed,
        cell_acc_rounding_mode,  # Output gate accumulator quant
        output_acc_scale,
        output_acc_zero_point,
        output_acc_bit_width,
        output_acc_narrow_range,
        output_acc_signed,
        output_acc_rounding_mode,  # Input gate sigmoid quant
        input_sigmoid_scale,
        input_sigmoid_zero_point,
        input_sigmoid_bit_width,
        input_sigmoid_narrow_range,
        input_sigmoid_signed,
        input_sigmoid_rounding_mode,  # Forget gate sigmoid quant
        forget_sigmoid_scale,
        forget_sigmoid_zero_point,
        forget_sigmoid_bit_width,
        forget_sigmoid_narrow_range,
        forget_sigmoid_signed,
        forget_sigmoid_rounding_mode,  # Cell gate tanh quant
        cell_tanh_scale,
        cell_tanh_zero_point,
        cell_tanh_bit_width,
        cell_tanh_narrow_range,
        cell_tanh_signed,
        cell_tanh_rounding_mode,  # Output gate sigmoid quant
        output_sigmoid_scale,
        output_sigmoid_zero_point,
        output_sigmoid_bit_width,
        output_sigmoid_narrow_range,
        output_sigmoid_signed,
        output_sigmoid_rounding_mode,  # Hidden state tanh quant
        hidden_state_tanh_scale,
        hidden_state_tanh_zero_point,
        hidden_state_tanh_bit_width,
        hidden_state_tanh_narrow_range,
        hidden_state_tanh_signed,
        hidden_state_tanh_rounding_mode,
    ):
        return g.op(
            f"{DOMAIN_STRING}::QuantLSTMCell",  # Tensors
            # Input values
            quant_input,
            quant_hidden_state,
            quant_cell_state,
            quant_weight_ii,
            quant_weight_if,
            quant_weight_ic,
            quant_weight_io,
            quant_weight_hi,
            quant_weight_hf,
            quant_weight_hc,
            quant_weight_ho,
            quant_bias_input,
            quant_bias_forget,
            quant_bias_cell,
            quant_bias_output,  # Output quant
            output_scale,
            output_zero_point,
            output_bit_width,  # Cell state quant
            cell_state_scale,
            cell_state_zero_point,
            cell_state_bit_width,  # Input gate accumulator quant
            input_acc_scale,
            input_acc_zero_point,
            input_acc_bit_width,  # Forget gate accumulator quant
            forget_acc_scale,
            forget_acc_zero_point,
            forget_acc_bit_width,  # Cell gate accumulator quant
            cell_acc_scale,
            cell_acc_zero_point,
            cell_acc_bit_width,  # Output gate accumulator quant
            output_acc_scale,
            output_acc_zero_point,
            output_acc_bit_width,  # Input gate sigmoid quant
            input_sigmoid_scale,
            input_sigmoid_zero_point,
            input_sigmoid_bit_width,  # Forget gate sigmoid quant
            forget_sigmoid_scale,
            forget_sigmoid_zero_point,
            forget_sigmoid_bit_width,  # Cell gate tanh quant
            cell_tanh_scale,
            cell_tanh_zero_point,
            cell_tanh_bit_width,  # Output gate sigmoid quant
            output_sigmoid_scale,
            output_sigmoid_zero_point,
            output_sigmoid_bit_width,  # Hidden state tanh quant
            hidden_state_tanh_scale,
            hidden_state_tanh_zero_point,
            hidden_state_tanh_bit_width,
            # Attributes
            batch_first_i=batch_first,
            reverse_input_i=reverse_input,
            cifg_i=cifg,
            output_narrow_i=output_narrow_range,
            output_signed_i=output_signed,
            output_rounding_mode_s=output_rounding_mode,
            cell_state_narrow_i=cell_state_narrow_range,
            cell_state_signed_i=cell_state_signed,
            cell_state_rounding_mode_s=cell_state_rounding_mode,
            input_acc_narrow_i=input_acc_narrow_range,
            input_acc_signed_i=input_acc_signed,
            input_acc_rounding_mode_s=input_acc_rounding_mode,
            forget_acc_narrow_i=forget_acc_narrow_range,
            forget_acc_signed_i=forget_acc_signed,
            forget_acc_rounding_mode_s=forget_acc_rounding_mode,
            cell_acc_narrow_i=cell_acc_narrow_range,
            cell_acc_signed_i=cell_acc_signed,
            cell_acc_rounding_mode_s=cell_acc_rounding_mode,
            output_acc_narrow_i=output_acc_narrow_range,
            output_acc_signed_i=output_acc_signed,
            output_acc_rounding_mode_s=output_acc_rounding_mode,
            input_sigmoid_narrow_i=input_sigmoid_narrow_range,
            input_sigmoid_signed_i=input_sigmoid_signed,
            input_sigmoid_rounding_mode_s=input_sigmoid_rounding_mode,
            forget_sigmoid_narrow_i=forget_sigmoid_narrow_range,
            forget_sigmoid_signed_i=forget_sigmoid_signed,
            forget_sigmoid_rounding_mode_s=forget_sigmoid_rounding_mode,
            cell_tanh_narrow_i=cell_tanh_narrow_range,
            cell_tanh_signed_i=cell_tanh_signed,
            cell_tanh_rounding_mode_s=cell_tanh_rounding_mode,
            output_sigmoid_narrow_range_i=output_sigmoid_narrow_range,
            output_sigmoid_signed_i=output_sigmoid_signed,
            output_sigmoid_rounding_mode_s=output_sigmoid_rounding_mode,
            hidden_state_tanh_narrow_i=hidden_state_tanh_narrow_range,
            hidden_state_tanh_signed_i=hidden_state_tanh_signed,
            hidden_state_tanh_rounding_mode_s=hidden_state_tanh_rounding_mode,
            # PyTorch requires to specify the number of outputs manually
            outputs=3,
        )

    @staticmethod
    def forward(
        ctx,  # args and kwargs passed from _QuantLSTMLayer
        quant_input,
        quant_hidden_state,
        quant_cell_state,
        quant_weight_ii,
        quant_weight_if,
        quant_weight_ic,
        quant_weight_io,
        quant_weight_hi,
        quant_weight_hf,
        quant_weight_hc,
        quant_weight_ho,
        quant_bias_input,
        quant_bias_forget,
        quant_bias_cell,
        quant_bias_output,  # Symbolic kwargs passed from BrevitasQuantLSTMLayerHandler
        batch_first,
        reverse_input,
        cifg,  # Output quant
        output_scale,
        output_zero_point,
        output_bit_width,
        output_narrow_range,
        output_signed,
        output_rounding_mode,  # Cell state quant
        cell_state_scale,
        cell_state_zero_point,
        cell_state_bit_width,
        cell_state_narrow_range,
        cell_state_signed,
        cell_state_rounding_mode,  # Input gate accumulator quant
        input_acc_scale,
        input_acc_zero_point,
        input_acc_bit_width,
        input_acc_narrow_range,
        input_acc_signed,
        input_acc_rounding_mode,  # Forget gate accumulator quant
        forget_acc_scale,
        forget_acc_zero_point,
        forget_acc_bit_width,
        forget_acc_narrow_range,
        forget_acc_signed,
        forget_acc_rounding_mode,  # Cell gate accumulator quant
        cell_acc_scale,
        cell_acc_zero_point,
        cell_acc_bit_width,
        cell_acc_narrow_range,
        cell_acc_signed,
        cell_acc_rounding_mode,  # Output gate accumulator quant
        output_acc_scale,
        output_acc_zero_point,
        output_acc_bit_width,
        output_acc_narrow_range,
        output_acc_signed,
        output_acc_rounding_mode,  # Input gate sigmoid quant
        input_sigmoid_scale,
        input_sigmoid_zero_point,
        input_sigmoid_bit_width,
        input_sigmoid_narrow_range,
        input_sigmoid_signed,
        input_sigmoid_rounding_mode,  # Forget gate sigmoid quant
        forget_sigmoid_scale,
        forget_sigmoid_zero_point,
        forget_sigmoid_bit_width,
        forget_sigmoid_narrow_range,
        forget_sigmoid_signed,
        forget_sigmoid_rounding_mode,  # Cell gate tanh quant
        cell_tanh_scale,
        cell_tanh_zero_point,
        cell_tanh_bit_width,
        cell_tanh_narrow_range,
        cell_tanh_signed,
        cell_tanh_rounding_mode,  # Output gate sigmoid quant
        output_sigmoid_scale,
        output_sigmoid_zero_point,
        output_sigmoid_bit_width,
        output_sigmoid_narrow_range,
        output_sigmoid_signed,
        output_sigmoid_rounding_mode,  # Hidden state tanh quant
        hidden_state_tanh_scale,
        hidden_state_tanh_zero_point,
        hidden_state_tanh_bit_width,
        hidden_state_tanh_narrow_range,
        hidden_state_tanh_signed,
        hidden_state_tanh_rounding_mode,
    ):
        # Tp simplify things, here we are returning the outputs
        # as if they were already concatenated. Scale/zp/bw are avoided too.
        # This preserves output shapes but not values.
        # See _QuantLSTMCell for the actual implementation.
        quant_outputs = torch.zeros(
            quant_input.size(0), quant_input.size(1), quant_hidden_state.size(1), device=quant_hidden_state.device
        )
        return quant_outputs, quant_hidden_state, quant_cell_state



def reseed(newseed):
    import numpy
    import onnxruntime
    import tensorflow

    print(f"pytest-randomly: reseed with {newseed}")
    onnxruntime.set_seed(newseed)
    tensorflow.random.set_seed(newseed)
    numpy.random.seed(seed=newseed)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize
import numpy as np
import onnxruntime as rt
from tqdm import tqdm

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.core.onnx_exec import execute_onnx
from qonnx.transformation.change_batchsize import ChangeBatchSize
from qonnx.transformation.expose_intermediate import ExposeIntermediateTensorsPatternList
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.util.onnx import valueinfo_to_tensor

OUTPUT_MODE_IND = "tensor_index"
OUTPUT_MODE_NAME = "tensor_name"

output_modes = {OUTPUT_MODE_IND, OUTPUT_MODE_NAME}

output_mode_options = clize.parameters.mapped(
    [
        (OUTPUT_MODE_IND, [OUTPUT_MODE_IND], "Output files named by index"),
        (OUTPUT_MODE_NAME, [OUTPUT_MODE_NAME], "Output files named by tensor"),
    ]
)


def exec_qonnx(
    qonnx_model_file,
    *in_npy,
    override_batchsize: int = None,
    override_opset: int = None,
    expose_intermediates: str = None,
    output_prefix: str = "out_",
    output_mode: output_mode_options = OUTPUT_MODE_NAME,
    argmax_verify_npy: str = None,
    save_modified_model: str = None,
    input_pix2float=False,
    input_zerocenter=False,
    maxiters: int = None,
    output_nosave=False
):
    """Execute a given QONNX model by initializing its inputs from .npy files, and write outputs
    as .npy files.
    The input model have been previously cleaned by the cleanup transformation or commandline tool.

    :param qonnx_model_file: Filename for the input ONNX model
    :param in_npy: List of .npy files to supply as inputs. If not specified, inputs will be set to zero.
    :param override_batchsize: If specified, override the batch size for the ONNX graph
    :param override_opset: If specified, override the imported ONNX opset to this version.
    :param expose_intermediates: Comma-separated list of tensor name patterns.
        Matched patterns will expose intermediate outputs as top-level outputs.
    :param output_prefix: Prefix for the generated output files.
    :param output_mode: Naming mode for generated output files.
    :param argmax_verify_npy: If specified, take argmax of output and compare to this file for top-1 accuracy measurement
    :param save_modified_model: If specified, save the modified model
        (after batchsize changes or exposed intermediate tensors) with this filename
    :param input_pix2float: If specified, do uint8 [0,255] -> fp32 [0,1] mapping for input
    :param input_zerocenter: If specified together with pix2float, do uint8 [0,255] -> fp32 [-1,+1] mapping for input
    :param maxiters: If specified, limit maximum number of iterations (batches) to be processed
    :param output_nosave: If specified, do not save output tensors to files
    """
    assert output_mode in output_modes, "Unrecognized output mode"

    # Execute transformation
    model = ModelWrapper(qonnx_model_file)
    if override_opset is not None:
        model.model.opset_import[0].version = override_opset

    if override_batchsize is not None:
        model = model.transform(ChangeBatchSize(override_batchsize))
        model = model.transform(InferShapes())
        bsize = override_batchsize
    else:
        bsize = model.get_tensor_shape(model.graph.input[0].name)[0]

    if expose_intermediates is not None:
        pattern_list = expose_intermediates.split(",")
        pattern_list = [x.strip() for x in pattern_list]
        model = model.transform(FoldConstants(exclude_op_types=[]))
        model = model.transform(ExposeIntermediateTensorsPatternList(pattern_list, dynamic_only=True))

    if save_modified_model is not None:
        model.save(save_modified_model)

    n_custom_nodes = len(model.get_finn_nodes())
    if n_custom_nodes == 0:
        print("No custom qonnx nodes found, running in onnxruntime")

    ok = 0
    nok = 0
    iter = 0
    dset_size = 0
    n_dset_iters = 0
    inp_data = []
    labels = None
    if len(in_npy) > 0:
        # load provided npy files and arrange in batches
        inp_data = [np.load(x) for x in in_npy]
        inp_data_reshaped = []
        for inp in inp_data:
            dset_size = inp.shape[0]
            assert dset_size % bsize == 0, "Batch size %d must divide dataset size %d" % (bsize, dset_size)
            n_dset_iters = dset_size // bsize
            inp = inp.reshape(n_dset_iters, bsize, *inp.shape[1:])
            inp_data_reshaped.append(inp)
        inp_data = inp_data_reshaped
        if argmax_verify_npy is not None:
            labels = np.load(argmax_verify_npy)
            assert labels.shape[0] == dset_size, "Label size must match dataset size"
            labels = labels.reshape(n_dset_iters, bsize, *labels.shape[1:])
    else:
        # create 0-valued tensors of appropriate shape
        n_dset_iters = 1
        inp_data = [valueinfo_to_tensor(model.get_tensor_valueinfo(i.name)) for i in model.graph.input]
        inp_data = [np.expand_dims(x, axis=0) for x in inp_data]

    if maxiters is not None:
        n_dset_iters = min(n_dset_iters, maxiters)

    pbar = tqdm(range(n_dset_iters))

    for iter in pbar:
        iter_suffix = "_batch%d" % iter
        idict = {}
        if not argmax_verify_npy:
            pbar.set_description("Batch [%d/%d]: running" % (iter + 1, n_dset_iters))
        # supply inputs and execute
        for inp_ind, inp in enumerate(model.graph.input):
            if input_pix2float:
                idict[inp.name] = (inp_data[inp_ind][iter] / 255.0).astype(np.float32)
                if input_zerocenter:
                    idict[inp.name] = (2 * idict[inp.name] - 1.0).astype(np.float32)
            else:
                idict[inp.name] = inp_data[inp_ind][iter]
        if n_custom_nodes > 0:
            # run node-by-node in qonnx
            odict = execute_onnx(model, idict)
        else:
            # run using onnxruntime
            sess = rt.InferenceSession(model.model.SerializeToString())
            output_list = sess.run(None, idict)
            odict = {outp.name: output_list[oind] for oind, outp in enumerate(model.graph.output)}
        if not output_nosave:
            for out_ind, outp in enumerate(model.graph.output):
                # save generated outputs
                if output_mode == OUTPUT_MODE_IND:
                    oname = "%d" % out_ind
                elif output_mode == OUTPUT_MODE_NAME:
                    oname = outp.name
                np.save(output_prefix + oname + iter_suffix + ".npy", odict[outp.name])
        if argmax_verify_npy:
            # measure accuracy for output
            ret = odict[model.graph.output[0].name]
            ret = np.argmax(ret, axis=-1)
            ok_batch = np.count_nonzero(ret == labels[iter])
            nok_batch = bsize - ok_batch
            ok += ok_batch
            nok += nok_batch
            accuracy_batch = ok_batch / bsize
            accuracy_overall = ok / (ok + nok)
            pbar.set_description(
                "Batch [%d/%d]: ok %d nok %d accuracy %f (overall ok %d nok %d accuracy %f)"
                % (iter + 1, n_dset_iters, ok_batch, nok_batch, accuracy_batch, ok, nok, accuracy_overall)
            )


def main():
    clize.run(exec_qonnx)


if __name__ == "__main__":
    main()



import onnx
import tensorflow as tf
import tf2onnx
from collections import OrderedDict
from qkeras.utils import REGISTERED_LAYERS as QKERAS_LAYERS

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.util.cleanup import cleanup_model

from .qkeras.onnx import get_qkeras_onnx_handlers
from .qkeras.qlayers import extract_quantizers_from_layer

_unsupported_layers = [
    # These require some extra work
    "QBatchNormalization",
    "QConv2DBatchnorm",
    "QDepthwiseConv2DBatchnorm",
]

# Skip remove_identity optimizer
del tf2onnx.optimizer._optimizers["remove_identity"]


def add_value_info_for_constants(model: onnx.ModelProto):
    """
    Currently onnx.shape_inference doesn't use the shape of initializers, so add
    that info explicitly as ValueInfoProtos.
    Mutates the model.
    Args:
        model: The ModelProto to update.
    """
    # All (top-level) constants will have ValueInfos before IRv4 as they are all inputs
    if model.ir_version < 4:
        return model

    def add_const_value_infos_to_graph(graph: onnx.GraphProto):
        inputs = {i.name for i in graph.input}
        existing_info = {vi.name: vi for vi in graph.value_info}
        for init in graph.initializer:
            # Check it really is a constant, not an input
            if init.name in inputs:
                continue

            # The details we want to add
            elem_type = init.data_type
            shape = init.dims

            # Get existing or create new value info for this constant
            vi = existing_info.get(init.name)
            if vi is None:
                vi = graph.value_info.add()
                vi.name = init.name

            # Even though it would be weird, we will not overwrite info even if it doesn't match
            tt = vi.type.tensor_type
            if tt.elem_type == onnx.TensorProto.UNDEFINED:
                tt.elem_type = elem_type
            if not tt.HasField("shape"):
                # Ensure we set an empty list if the const is scalar (zero dims)
                tt.shape.dim.extend([])
                for dim in shape:
                    tt.shape.dim.add().dim_value = dim

        # Handle subgraphs
        for node in graph.node:
            for attr in node.attribute:
                # Ref attrs refer to other attrs, so we don't need to do anything
                if attr.ref_attr_name != "":
                    continue

                if attr.type == onnx.AttributeProto.GRAPH:
                    add_const_value_infos_to_graph(attr.g)
                if attr.type == onnx.AttributeProto.GRAPHS:
                    for g in attr.graphs:
                        add_const_value_infos_to_graph(g)

    add_const_value_infos_to_graph(model.graph)
    return model


def _is_qkeras_model(model):
    """Check if the model has any qkeras layers, so we can handle the qkeras layers separately

    Args:
        model: the model we want to convert

    Returns:
        True if the model contains any qkeras layer
    """

    def iterate_model(model):
        for layer in model.layers:
            if isinstance(layer, tf.keras.Model):
                found_qkeras = iterate_model(layer)
                if found_qkeras:
                    return True
            elif layer.__class__.__name__ in QKERAS_LAYERS:
                return True

        return False

    return iterate_model(model)


def _check_supported_layers(model):
    """Check if all the layers in the model are supported for conversion

    Args:
        model: the tf.keras model we want to convert

    Returns:
        Exception if an unsupported layer is found in the model
    """

    def iterate_model(model):
        for layer in model.layers:
            if isinstance(layer, tf.keras.Model):
                iterate_model(layer)
            elif layer.__class__.__name__ in _unsupported_layers:
                raise Exception("Currently unsupported layer found in QKeras model: {}".format(layer.__class__.__name__))

    iterate_model(model)


def _strip_qkeras_model(model):
    """Strip a qkeras model to obtain the keras model and obtain the quant nodes.

    Args:
        model: the tf.keras model we want to convert

    Returns:
        The stripped model, and the quantizers in a dictionary format
    """
    quantizers = OrderedDict()

    def extract_quantizers(layer):
        keras_cls_name, layer_cfg, layer_quantizers = extract_quantizers_from_layer(layer)
        if layer_quantizers:
            layer_quantizers = {
                k: None if v == "None" else v for k, v in layer_quantizers.items()
            }  # Get rid of 'None' strings
            layer_quantizers["input"] = layer.input.name
            quantizers[layer.name] = layer_quantizers

        layer_class = tf.keras.layers.__dict__.get(keras_cls_name, None)
        if layer_class is None:
            raise Exception("Cannot create Keras layer from QKeras class {}".format(keras_cls_name))

        return layer_class.from_config(layer_cfg)

    stripped_model = tf.keras.models.clone_model(model, clone_function=extract_quantizers)
    stripped_model.set_weights(model.get_weights())
    return stripped_model, quantizers


# tests run without this function
def _convert_quantizers_to_nodes(onnx_model, quantizers_dict):
    for node_name, quantizers in quantizers_dict.items():
        print(node_name, quantizers)

    for n in onnx_model.graph.node:
        print(n)

    return onnx_model.model


def from_keras(
    model,
    name="qkeras_to_qonnx_converted",
    input_signature=None,
    opset=None,
    custom_ops=None,
    custom_op_handlers=None,
    custom_rewriter=None,
    inputs_as_nchw=None,
    extra_opset=None,
    shape_override=None,
    target=None,
    large_model=False,
    output_path=None,
):
    """Convert a keras model to QONNX. The API follows the `from_keras` function of tf2onnx.

    Args:
        model: the tf.keras model we want to convert
        input_signature: a tf.TensorSpec or a numpy array defining the shape/dtype of the input
        opset: the opset to be used for the ONNX model, default is the latest
        custom_ops: if a model contains ops not recognized by onnx runtime,
            you can tag these ops with a custom op domain so that the
            runtime can still open the model. Type is a dictionary `{op name: domain}`.
        target: list of workarounds applied to help certain platforms
        custom_op_handlers: dictionary of custom ops handlers
        custom_rewriter: list of custom graph rewriters
        extra_opset: list of extra opset's, for example the opset's used by custom ops
        shape_override: dict with inputs that override the shapes given by tensorflow
        inputs_as_nchw: transpose inputs in list from nchw to nhwc
        large_model: use the ONNX external tensor storage format
        output_path: save model to output_path

    Returns:
        An ONNX model_proto and an external_tensor_storage dict.
    """

    assert not large_model  # TODO for now, let's focus only on models that don't store tensors externally

    if _is_qkeras_model(model):
        _check_supported_layers(model)
        keras_model, quantizers = _strip_qkeras_model(model)
    else:
        keras_model, quantizers = model, {}

    qkeras_op_handlers = get_qkeras_onnx_handlers(quantizers)

    if custom_op_handlers is not None:
        qkeras_op_handlers.update(custom_op_handlers)

    model_proto, external_storage = tf2onnx.convert.from_keras(
        keras_model,
        input_signature=input_signature,
        opset=opset,
        custom_ops=custom_ops,
        custom_op_handlers=qkeras_op_handlers,
        custom_rewriter=custom_rewriter,
        inputs_as_nchw=inputs_as_nchw,
        extra_opset=extra_opset,
        shape_override=shape_override,
        target=target,
        large_model=large_model,
        output_path=None,
    )

    onnx_model = ModelWrapper(model_proto)
    # Set the first value of input/output shape to 1, currently this is set to unknown,
    # because it is technically the batch size
    if not (len(onnx_model.graph.input) == 1 and len(onnx_model.graph.output) == 1):
        raise ValueError("Qkeras to QONNX conversion only supports models with exactly one input and output.")
    inp_shape = onnx_model.get_tensor_shape(onnx_model.graph.input[0].name)
    out_shape = onnx_model.get_tensor_shape(onnx_model.graph.output[0].name)
    inp_shape[0] = 1
    out_shape[0] = 1
    onnx_model.set_tensor_shape(onnx_model.graph.input[0].name, inp_shape)
    onnx_model.set_tensor_shape(onnx_model.graph.output[0].name, out_shape)

    # Set all Quant output tensors to float32 datatype, otherwise they are undefined and crash ONNX execution
    qonnx_domain_ops = ["Quant", "Trunc", "BipolarQuant"]
    for q_op_type in qonnx_domain_ops:
        quant_nodes = onnx_model.get_nodes_by_op_type(q_op_type)
        q_node_outputs = [qn.output[0] for qn in quant_nodes]
        for tensor in onnx_model.graph.value_info:
            if tensor.name in q_node_outputs:
                tensor.type.tensor_type.elem_type = 1

    onnx_model = cleanup_model(onnx_model)
    onnx_model.model = add_value_info_for_constants(onnx_model.model)

    if output_path is not None:
        onnx_model.save(output_path)

    return onnx_model.model, external_storage



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

import numpy as np
import os
import urllib.request

import qonnx.core.onnx_exec as oxe
from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.qcdq_to_qonnx import QCDQToQuant
from qonnx.util.cleanup import cleanup_model

model_details = {
    "MobileNetv2-w8a8": {
        "url": (
            "https://github.com/onnx/models/raw/main/validated/vision/classification/mobilenet/model/mobilenetv2-12-qdq.onnx"
        ),
        "input_shape": (1, 3, 224, 224),
        "input_range": (-1, +1),
        "exp_q_nodes": 171,
    },
}


def download_model(test_model):
    qonnx_url = model_details[test_model]["url"]
    # download test data
    dl_dir = "/tmp"
    dl_file = dl_dir + f"/{test_model}.onnx"
    urllib.request.urlretrieve(qonnx_url, dl_file)
    return dl_file


def get_golden_in_and_output(model, test_model):
    rng = np.random.RandomState(42)
    input_shape = model_details[test_model]["input_shape"]
    (low, high) = model_details[test_model]["input_range"]
    size = np.prod(np.asarray(input_shape))
    input_tensor = rng.uniform(low=low, high=high, size=size)
    input_tensor = input_tensor.astype(np.float32)
    input_tensor = input_tensor.reshape(input_shape)
    # use batch dim of 1 where needed (appears as 0-valued dim)
    for tensor_name in model.get_all_tensor_names():
        ts = model.get_tensor_shape(tensor_name)
        if len(ts) > 0 and ts[0] == 0:
            ts = list(ts)
            ts[0] = 1
            model.set_tensor_shape(tensor_name, ts)
    model.set_tensor_shape(model.graph.input[0].name, input_shape)
    input_dict = {model.graph.input[0].name: input_tensor}
    golden_output_dict = oxe.execute_onnx(model, input_dict)
    golden_result = golden_output_dict[model.graph.output[0].name]
    return input_tensor, golden_result


@pytest.mark.parametrize("test_model", model_details.keys())
def test_qcdq_to_qonnx(test_model):
    test_details = model_details[test_model]
    dl_file = download_model(test_model=test_model)
    assert os.path.isfile(dl_file)
    model = ModelWrapper(dl_file)
    model = cleanup_model(model)
    input_tensor, golden_result = get_golden_in_and_output(model, test_model)
    model = model.transform(QCDQToQuant())
    assert len(model.get_nodes_by_op_type("Quant")) == test_details["exp_q_nodes"]
    assert len(model.get_nodes_by_op_type("QuantizeLinear")) == 0
    assert len(model.get_nodes_by_op_type("DequantizeLinear")) == 0
    if test_model == "MobileNetv2-w8a8":
        pytest.xfail("MNv2 known to have off-by-one difference in inputs to Conv_30")
    model = cleanup_model(model)
    input_dict = {model.graph.input[0].name: input_tensor}
    produced_output_dict = oxe.execute_onnx(model, input_dict)
    produced_result = produced_output_dict[model.graph.output[0].name]
    assert np.isclose(golden_result, produced_result).all()



# Copyright (c) 2022, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import onnx.helper as oh
from pkgutil import get_data

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import NodeLocalTransformation
from qonnx.util.basic import get_by_name


class SumAndAnnotateConvWeights(NodeLocalTransformation):
    def applyNodeLocal(self, node):
        if node.op_type == "Conv":
            # read conv weight tensor from model
            W = self.ref_input_model.get_initializer(node.input[1])
            sum = float(np.sum(W))
            # add a dummy attribute for verification
            attr_proto = oh.make_attribute("conv_w_sum", sum)
            node.attribute.append(attr_proto)
        return (node, False)


def test_nodelocal_transform():
    # load the onnx model
    raw_m = get_data("qonnx.data", "onnx/mnist-conv/model.onnx")
    model = ModelWrapper(raw_m)
    model = model.transform(SumAndAnnotateConvWeights())
    for conv_node in model.get_nodes_by_op_type("Conv"):
        wsum = float(np.sum(model.get_initializer(conv_node.input[1])))
        assert get_by_name(conv_node.attribute, "conv_w_sum").f == wsum
